# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Top-module metadata."""

try:
    from ._version import __version__
except ImportError:
    __version__ = '0+unknown'



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Baby brain extraction from T2w images."""

from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.engine import tag


@tag('anat.brain-extraction')
def init_infant_brain_extraction_wf(
    *,
    omp_nthreads: int,
    sloppy: bool,
    age_months: int | None = None,
    ants_affine_init: bool = True,
    bspline_fitting_distance: int = 200,
    skull_strip_template: str = 'UNCInfant',
    template_specs: dict | None = None,
    mem_gb: float = 3.0,
    debug: bool = False,
    name: str = 'infant_brain_extraction_wf',
):
    """
    Build an atlas-based brain extraction pipeline for infant T2w MRI data.

    Pros/Cons of available templates
    --------------------------------
    * MNIInfant
     + More cohorts available for finer-grain control
     + T1w/T2w images available
     - Template masks are poor

    * UNCInfant
     + Accurate masks
     - No T2w image available


    Parameters
    ----------
    age_months : :obj:`int`
        Age of this participant, in months.
    ants_affine_init : :obj:`bool`, optional
        Set-up a pre-initialization step with ``antsAI`` to account for mis-oriented images.
    bspline_fitting_distance : :obj:`float`
        Distance in mm between B-Spline control points for N4 INU estimation.
    sloppy : :obj:`bool`
        Run in *sloppy* mode.
    skull_strip_template : :obj:`str`
        A TemplateFlow ID indicating which template will be used as target for atlas-based
        segmentation.
    template_specs : :obj:`dict`
        Additional template specifications (e.g., resolution or cohort) to correctly select
        the adequate template instance.
    mem_gb : :obj:`float`
        Base memory fingerprint unit.
    name : :obj:`str`
        This particular workflow's unique name (Nipype requirement).
    omp_nthreads : :obj:`int`
        The number of threads for individual processes in this workflow.
    debug : :obj:`bool`
        Produce intermediate registration files

    Inputs
    ------
    t2w_preproc : :obj:`str`
        The preprocessed T2w image (Denoising/INU/N4)

    Outputs
    -------
    t2w_preproc : :obj:`str`
        The preprocessed T2w image (INU and clipping).
    t2w_brain : :obj:`str`
        The preprocessed, brain-extracted T2w image.
    out_mask : :obj:`str`
        The brainmask projected from the template into the T2w, after
        binarization.
    out_probmap : :obj:`str`
        The same as above, before binarization.

    """
    from nipype.interfaces.ants import ImageMath, N4BiasFieldCorrection
    from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
    from niworkflows.interfaces.fixes import FixHeaderRegistration as Registration
    from niworkflows.interfaces.nibabel import (
        ApplyMask,
        Binarize,
        BinaryDilation,
        IntensityClip,
    )

    from nibabies.data import load as load_data
    from nibabies.utils.misc import cohort_by_months

    # handle template specifics
    template_specs = template_specs or {}
    if skull_strip_template == 'MNIInfant':
        template_specs['resolution'] = 2 if sloppy else 1

    if not template_specs.get('cohort'):
        if age_months is None:
            raise KeyError(f'Age or cohort for {skull_strip_template} must be provided!')
        template_specs['cohort'] = cohort_by_months(skull_strip_template, age_months)

    template_files = fetch_templates(skull_strip_template, template_specs)

    # main workflow
    workflow = pe.Workflow(name)

    inputnode = pe.Node(niu.IdentityInterface(fields=['t2w_preproc']), name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['t2w_preproc', 't2w_brain', 'out_mask', 'out_probmap']),
        name='outputnode',
    )

    # Ensure template comes with a range of intensities ANTs will like
    clip_tmpl = pe.Node(IntensityClip(p_max=99), name='clip_tmpl')
    clip_tmpl.inputs.in_file = _pop(template_files['anat'])

    # Generate laplacian registration targets
    lap_tmpl = pe.Node(ImageMath(operation='Laplacian', op2='0.4 1'), name='lap_tmpl')
    lap_t2w = pe.Node(ImageMath(operation='Laplacian', op2='0.4 1'), name='lap_t2w')
    norm_lap_tmpl = pe.Node(niu.Function(function=_norm_lap), name='norm_lap_tmpl')
    norm_lap_t2w = pe.Node(niu.Function(function=_norm_lap), name='norm_lap_t2w')

    # Merge image nodes
    mrg_tmpl = pe.Node(niu.Merge(2), name='mrg_tmpl', run_without_submitting=True)
    mrg_t2w = pe.Node(niu.Merge(2), name='mrg_t2w', run_without_submitting=True)
    bin_regmask = pe.Node(Binarize(thresh_low=0.20), name='bin_regmask')
    bin_regmask.inputs.in_file = str(template_files['mask'])
    refine_mask = pe.Node(BinaryDilation(radius=3, iterations=2), name='refine_mask')

    fixed_masks = pe.Node(niu.Merge(4), name='fixed_masks', run_without_submitting=True)
    fixed_masks.inputs.in1 = 'NULL'
    fixed_masks.inputs.in2 = 'NULL'
    fixed_masks.inputs.in3 = (
        'NULL' if not template_files['regmask'] else _pop(template_files['regmask'])
    )

    # Set up initial spatial normalization
    ants_params = 'testing' if sloppy else 'precise'
    norm = pe.Node(
        Registration(from_file=load_data(f'antsBrainExtraction_{ants_params}.json')),
        name='norm',
        n_procs=omp_nthreads,
        mem_gb=mem_gb,
    )
    norm.inputs.float = sloppy
    if debug:
        norm.inputs.args = '--write-interval-volumes 5'

    map_mask_t2w = pe.Node(
        ApplyTransforms(interpolation='Gaussian', float=True),
        name='map_mask_t2w',
        mem_gb=1,
    )

    # map template brainmask to t2w space
    map_mask_t2w.inputs.input_image = str(template_files['mask'])

    thr_t2w_mask = pe.Node(Binarize(thresh_low=0.80), name='thr_t2w_mask')

    # Refine INU correction
    final_n4 = pe.Node(
        N4BiasFieldCorrection(
            dimension=3,
            bspline_fitting_distance=bspline_fitting_distance,
            save_bias=True,
            copy_header=True,
            n_iterations=[50] * 5,
            convergence_threshold=1e-7,
            rescale_intensities=True,
            shrink_factor=4,
        ),
        n_procs=omp_nthreads,
        name='final_n4',
    )
    final_clip = pe.Node(IntensityClip(p_min=5.0, p_max=99.5), name='final_clip')
    apply_mask = pe.Node(ApplyMask(), name='apply_mask')

    # fmt:off
    workflow.connect([
        (inputnode, final_n4, [('t2w_preproc', 'input_image')]),
        # 1. Massage T2w
        (inputnode, mrg_t2w, [('t2w_preproc', 'in1')]),
        (inputnode, lap_t2w, [('t2w_preproc', 'op1')]),
        (inputnode, map_mask_t2w, [('t2w_preproc', 'reference_image')]),
        (bin_regmask, refine_mask, [('out_file', 'in_file')]),
        (refine_mask, fixed_masks, [('out_file', 'in4')]),
        (lap_t2w, norm_lap_t2w, [('output_image', 'in_file')]),
        (norm_lap_t2w, mrg_t2w, [('out', 'in2')]),
        # 2. Prepare template
        (clip_tmpl, lap_tmpl, [('out_file', 'op1')]),
        (lap_tmpl, norm_lap_tmpl, [('output_image', 'in_file')]),
        (clip_tmpl, mrg_tmpl, [('out_file', 'in1')]),
        (norm_lap_tmpl, mrg_tmpl, [('out', 'in2')]),
        # 3. Set normalization node inputs
        (mrg_tmpl, norm, [('out', 'fixed_image')]),
        (mrg_t2w, norm, [('out', 'moving_image')]),
        (fixed_masks, norm, [('out', 'fixed_image_masks')]),
        # 4. Map template brainmask into T2w space
        (norm, map_mask_t2w, [
            ('reverse_transforms', 'transforms'),
            ('reverse_invert_flags', 'invert_transform_flags')
        ]),
        (map_mask_t2w, thr_t2w_mask, [('output_image', 'in_file')]),
        (thr_t2w_mask, apply_mask, [('out_mask', 'in_mask')]),
        (final_n4, apply_mask, [('output_image', 'in_file')]),
        # 5. Refine T2w INU correction with brain mask
        (map_mask_t2w, final_n4, [('output_image', 'weight_image')]),
        (final_n4, final_clip, [('output_image', 'in_file')]),
        # 6. Outputs
        (final_clip, outputnode, [('out_file', 't2w_preproc')]),
        (map_mask_t2w, outputnode, [('output_image', 'out_probmap')]),
        (thr_t2w_mask, outputnode, [('out_mask', 'out_mask')]),
        (apply_mask, outputnode, [('out_file', 't2w_brain')]),
    ])
    # fmt:on

    if ants_affine_init:
        from nipype.interfaces.ants.utils import AI

        ants_kwargs = {
            'metric': ('Mattes', 32, 'Regular', 0.2),
            'transform': ('Affine', 0.1),
            'search_factor': (20, 0.12),
            'principal_axes': False,
            'convergence': (10, 1e-6, 10),
            'search_grid': (40, (0, 40, 40)),
            'verbose': True,
        }

        init_aff = pe.Node(
            AI(**ants_kwargs),
            name='init_aff',
            n_procs=omp_nthreads,
        )
        if template_files['regmask']:
            init_aff.inputs.fixed_image_mask = _pop(template_files['regmask'])

        # fmt:off
        workflow.connect([
            (clip_tmpl, init_aff, [('out_file', 'fixed_image')]),
            (inputnode, init_aff, [('t2w_preproc', 'moving_image')]),
            (init_aff, norm, [('output_transform', 'initial_moving_transform')]),
        ])
        # fmt:on

    return workflow


def _pop(in_files):
    if isinstance(in_files, list | tuple):
        return in_files[0]
    return in_files


def _norm_lap(in_file):
    from pathlib import Path

    import nibabel as nb
    import numpy as np
    from nipype.utils.filemanip import fname_presuffix

    img = nb.load(in_file)
    data = img.get_fdata()
    data -= np.median(data)
    l_max = np.percentile(data[data > 0], 99.8)
    l_min = np.percentile(data[data < 0], 0.2)
    data[data < 0] *= -1.0 / l_min
    data[data > 0] *= 1.0 / l_max
    data = np.clip(data, a_min=-1.0, a_max=1.0)

    out_file = fname_presuffix(
        Path(in_file).name, suffix='_norm', newpath=str(Path.cwd().absolute())
    )
    hdr = img.header.copy()
    hdr.set_data_dtype('float32')
    img.__class__(data.astype('float32'), img.affine, hdr).to_filename(out_file)
    return out_file


def fetch_templates(template: str, specs: dict) -> dict:
    from templateflow.api import get

    template_files = {}
    # Anatomical reference
    template_files['anat'] = get(template, suffix='T1w', desc=None, raise_empty=True, **specs)
    # Anatomical mask, prefer probseg if available
    template_files['mask'] = get(template, label='brain', suffix='probseg', **specs) or get(
        template, desc='brain', suffix='mask', **specs
    )
    # More dilated mask to facilitate registration
    template_files['regmask'] = get(
        template, label='BrainCerebellumExtraction', suffix='mask', **specs
    )
    return template_files



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Interfaces to generate reportlets."""

import logging
import os
import re
import time
from collections import Counter
from pathlib import Path

from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    Directory,
    File,
    InputMultiObject,
    SimpleInterface,
    Str,
    TraitedSpec,
    isdefined,
    traits,
)
from niworkflows.interfaces.reportlets import base as nrb

LOGGER = logging.getLogger('nipype.interface')

SUBJECT_TEMPLATE = """\
\t<ul class="elem-desc">
\t\t<li>Subject ID: {subject_id}</li>
\t\t<li>Session ID: {session_id}</li>
\t\t<li>Chronological age (months): {age}</li>
\t\t<li>Structural images: {num_t1w} T1-weighted, {num_t2w} T2-weighted</li>
\t\t<li>Anatomical reference space: {anat_ref}<li>
\t\t<li>Functional series: {n_bold:d}</li>
{tasks}
\t\t<li>Standard output spaces: {std_spaces}</li>
\t\t<li>Non-standard output spaces: {nstd_spaces}</li>
\t\t<li>Surface reconstruction method: {recon_method}</li>
\t\t<li>Surface reconstruction status: {recon_status}</li>
\t</ul>
"""

FUNCTIONAL_TEMPLATE = """\
\t\t<details open>
\t\t<summary>Summary</summary>
\t\t<ul class="elem-desc">
\t\t\t<li>Original orientation: {ornt}</li>
\t\t\t<li>Repetition time (TR): {tr:.03g}s</li>
\t\t\t<li>Phase-encoding (PE) direction: {pedir}</li>
\t\t\t<li>{multiecho}</li>
\t\t\t<li>Slice timing correction: {stc}</li>
\t\t\t<li>Susceptibility distortion correction: {sdc}</li>
\t\t\t<li>Registration: {registration}</li>
\t\t\t<li>Non-steady-state volumes: {dummy_scan_desc}</li>
\t\t</ul>
\t\t</details>
"""

ABOUT_TEMPLATE = """\t<ul>
\t\t<li>NiBabies version: {version}</li>
\t\t<li>NiBabies command: <code>{command}</code></li>
\t\t<li>Date preprocessed: {date}</li>
\t</ul>
</div>
"""


# TODO: Move to niworkflows
class SummaryOutputSpec(TraitedSpec):
    out_report = File(exists=True, desc='HTML segment containing summary')


class SummaryInterface(SimpleInterface):
    output_spec = SummaryOutputSpec

    def _run_interface(self, runtime):
        segment = self._generate_segment()
        fname = os.path.join(runtime.cwd, 'report.html')
        with open(fname, 'w') as fobj:
            fobj.write(segment)
        self._results['out_report'] = fname
        return runtime

    def _generate_segment(self):
        raise NotImplementedError


class SubjectSummaryInputSpec(BaseInterfaceInputSpec):
    t1w = InputMultiObject(File(exists=True), desc='T1w structural images')
    t2w = InputMultiObject(File(exists=True), desc='T2w structural images')
    subjects_dir = Directory(desc='Infant FreeSurfer subjects directory')
    subject_id = Str(mandatory=True, desc='Subject ID')
    session_id = Str(desc='Session ID')
    anatomical_reference = traits.Enum('T1w', 'T2w', mandatory=True)
    bold = InputMultiObject(
        traits.Either(File(exists=True), traits.List(File(exists=True))),
        desc='BOLD functional series',
    )
    std_spaces = traits.List(Str, desc='list of standard spaces')
    nstd_spaces = traits.List(Str, desc='list of non-standard spaces')
    recon_method = traits.Enum(
        None,
        'freesurfer',
        'infantfs',
        'mcribs',
        desc='surface reconstruction method',
    )
    age = traits.Int(desc='Chronological age in months at the time of session')


class SubjectSummaryOutputSpec(SummaryOutputSpec):
    # This exists to ensure that the summary is run prior to the first ReconAll
    # call, allowing a determination whether there is a pre-existing directory
    subject_id = Str(desc='Surface reconstruction subject ID')


class SubjectSummary(SummaryInterface):
    input_spec = SubjectSummaryInputSpec
    output_spec = SubjectSummaryOutputSpec

    def _run_interface(self, runtime):
        if self.inputs.subject_id:
            self._recon_id = f'sub-{self.inputs.subject_id}'
            if self.inputs.session_id:
                self._recon_id += f'_ses-{self.inputs.session_id}'
            self._results['subject_id'] = self._recon_id
        return super()._run_interface(runtime)

    def _generate_segment(self):
        BIDS_NAME = re.compile(
            r'^(.*\/)?'
            '(?P<subject_id>sub-[a-zA-Z0-9]+)'
            '(_(?P<session_id>ses-[a-zA-Z0-9]+))?'
            '(_(?P<task_id>task-[a-zA-Z0-9]+))?'
            '(_(?P<acq_id>acq-[a-zA-Z0-9]+))?'
            '(_(?P<rec_id>rec-[a-zA-Z0-9]+))?'
            '(_(?P<run_id>run-[a-zA-Z0-9]+))?'
        )

        recon_method = self.inputs.recon_method

        statuses = {'no': 'Not run', 'todo': 'Run by NiBabies', 'done': 'Pre-existing directory'}
        if not self.inputs.subjects_dir:
            recon_status = statuses['no']
        else:
            if recon_method == 'freesurfer':
                from smriprep.interfaces.freesurfer import ReconAll

                recon = ReconAll(
                    subjects_dir=self.inputs.subjects_dir,
                    subject_id=self._recon_id,
                    T1_files=self.inputs.t1w,
                    flags='-noskullstrip',
                )
                recon_status = (
                    statuses['done'] if recon.cmdline.startswith('echo') else statuses['todo']
                )

            elif recon_method == 'infantfs':
                from niworkflows.utils.connections import pop_file

                from nibabies.interfaces.freesurfer import InfantReconAll

                recon = InfantReconAll(
                    subjects_dir=self.inputs.subjects_dir,
                    subject_id=self._recon_id,
                    t1_file=pop_file(self.inputs.t1w),
                )
                recon_status = (
                    statuses['done'] if recon.cmdline.startswith('echo') else statuses['todo']
                )

            elif recon_method == 'mcribs':
                # Use fingerprint logfile to identify "MCRIBS" runs vs FreeSurfer
                fingerprint = (
                    Path(self.inputs.subjects_dir) / self._recon_id / 'scripts' / 'mcribs.log'
                )
                recon_status = statuses['done'] if fingerprint.exists() else statuses['todo']

        num_t1w, num_t2w = 0, 0
        if self.inputs.t1w:
            num_t1w = len(self.inputs.t1w)
        if self.inputs.t2w:
            num_t2w = len(self.inputs.t2w)

        # Add list of tasks with number of runs
        bold_series = self.inputs.bold if isdefined(self.inputs.bold) else []
        bold_series = [s[0] if isinstance(s, list) else s for s in bold_series]

        counts = Counter(
            BIDS_NAME.search(series).groupdict()['task_id'][5:] for series in bold_series
        )

        tasks = ''
        if counts:
            header = '\t\t<ul class="elem-desc">'
            footer = '\t\t</ul>'
            lines = [
                '\t\t\t<li>Task: {task_id} ({n_runs:d} run{s})</li>'.format(
                    task_id=task_id, n_runs=n_runs, s='' if n_runs == 1 else 's'
                )
                for task_id, n_runs in sorted(counts.items())
            ]
            tasks = '\n'.join([header] + lines + [footer])

        return SUBJECT_TEMPLATE.format(
            subject_id=self.inputs.subject_id,
            session_id=self.inputs.session_id,
            age=self.inputs.age,
            num_t1w=num_t1w,
            num_t2w=num_t2w,
            anat_ref=self.inputs.anatomical_reference,
            n_bold=len(bold_series),
            tasks=tasks,
            std_spaces=', '.join(self.inputs.std_spaces),
            nstd_spaces=', '.join(self.inputs.nstd_spaces),
            recon_method=recon_method,
            recon_status=recon_status,
        )


class FunctionalSummaryInputSpec(TraitedSpec):
    slice_timing = traits.Enum(False, True, 'TooShort', desc='Slice timing correction used')
    distortion_correction = traits.Str(
        desc='Susceptibility distortion correction method', mandatory=True
    )
    pe_direction = traits.Enum(
        None,
        'i',
        'i-',
        'j',
        'j-',
        'k',
        'k-',
        mandatory=True,
        desc='Phase-encoding direction detected',
    )
    registration = traits.Enum(
        'FSL', 'FreeSurfer', mandatory=True, desc='Functional/anatomical registration method'
    )
    fallback = traits.Bool(desc='Boundary-based registration rejected')
    registration_dof = traits.Enum(
        6, 9, 12, desc='Registration degrees of freedom', mandatory=True
    )
    registration_init = traits.Enum(
        't1w',
        't2w',
        'header',
        mandatory=True,
        desc='Whether to initialize registration with the "header"'
        ' or by centering the volumes ("t1w" or "t2w")',
    )
    tr = traits.Float(desc='Repetition time', mandatory=True)
    dummy_scans = traits.Either(traits.Int(), None, desc='number of dummy scans specified by user')
    algo_dummy_scans = traits.Int(desc='number of dummy scans determined by algorithm')
    echo_idx = InputMultiObject(traits.Str, usedefault=True, desc='BIDS echo identifiers')
    orientation = traits.Str(mandatory=True, desc='Orientation of the voxel axes')


class FunctionalSummary(SummaryInterface):
    input_spec = FunctionalSummaryInputSpec

    def _generate_segment(self):
        dof = self.inputs.registration_dof
        if isdefined(self.inputs.slice_timing):
            stc = {
                True: 'Applied',
                False: 'Not applied',
                'TooShort': 'Skipped (too few volumes)',
            }[self.inputs.slice_timing]
        else:
            stc = 'n/a'
        # TODO: Add a note about registration_init below?
        reg = {
            'FSL': [
                'FSL <code>flirt</code> with boundary-based registration'
                f' (BBR) metric - {dof} dof',
                'FSL <code>flirt</code> rigid registration - 6 dof',
            ],
            'FreeSurfer': [
                'FreeSurfer <code>bbregister</code> '
                f'(boundary-based registration, BBR) - {dof} dof',
                f'FreeSurfer <code>mri_coreg</code> - {dof} dof',
            ],
        }[self.inputs.registration][self.inputs.fallback]

        pedir = get_world_pedir(self.inputs.orientation, self.inputs.pe_direction)

        dummy_scan_tmp = '{n_dum}'
        if self.inputs.dummy_scans == self.inputs.algo_dummy_scans:
            dummy_scan_msg = ' '.join(
                [dummy_scan_tmp, '(Confirmed: {n_alg} automatically detected)']
            ).format(n_dum=self.inputs.dummy_scans, n_alg=self.inputs.algo_dummy_scans)
        # the number of dummy scans was specified by the user and
        # it is not equal to the number detected by the algorithm
        elif self.inputs.dummy_scans is not None:
            dummy_scan_msg = ' '.join(
                [dummy_scan_tmp, '(Warning: {n_alg} automatically detected)']
            ).format(n_dum=self.inputs.dummy_scans, n_alg=self.inputs.algo_dummy_scans)
        # the number of dummy scans was not specified by the user
        else:
            dummy_scan_msg = dummy_scan_tmp.format(n_dum=self.inputs.algo_dummy_scans)

        multiecho = 'Single-echo EPI sequence.'
        n_echos = len(self.inputs.echo_idx)
        if n_echos == 1:
            multiecho = (
                f'Multi-echo EPI sequence: only echo {self.inputs.echo_idx[0]} processed '
                'in single-echo mode.'
            )
        if n_echos > 2:
            multiecho = f'Multi-echo EPI sequence: {n_echos} echoes.'

        return FUNCTIONAL_TEMPLATE.format(
            pedir=pedir,
            stc=stc,
            sdc=self.inputs.distortion_correction,
            registration=reg,
            tr=self.inputs.tr,
            dummy_scan_desc=dummy_scan_msg,
            multiecho=multiecho,
            ornt=self.inputs.orientation,
        )


class AboutSummaryInputSpec(BaseInterfaceInputSpec):
    version = Str(desc='NiBabies version')
    command = Str(desc='NiBabies command')
    # Date not included - update timestamp only if version or command changes


class AboutSummary(SummaryInterface):
    input_spec = AboutSummaryInputSpec

    def _generate_segment(self):
        return ABOUT_TEMPLATE.format(
            version=self.inputs.version,
            command=self.inputs.command,
            date=time.strftime('%Y-%m-%d %H:%M:%S %z'),
        )


class LabeledHistogramInputSpec(nrb._SVGReportCapableInputSpec):
    in_file = traits.File(exists=True, mandatory=True, desc='Image containing values to plot')
    label_file = traits.File(
        exists=True,
        desc='Mask or label image where non-zero values will be used to extract data from in_file',
    )
    mapping = traits.Dict(desc='Map integer label values onto names of voxels')
    xlabel = traits.Str('voxels', usedefault=True, desc='Description of values plotted')


class LabeledHistogram(nrb.ReportingInterface):
    input_spec = LabeledHistogramInputSpec

    def _generate_report(self):
        import nibabel as nb
        import numpy as np
        import seaborn as sns
        from matplotlib import pyplot as plt
        from nilearn.image import resample_to_img

        report_file = self._out_report
        img = nb.load(self.inputs.in_file)
        data = img.get_fdata(dtype=np.float32)

        if self.inputs.label_file:
            label_img = nb.load(self.inputs.label_file)
            if label_img.shape != img.shape[:3] or not np.allclose(label_img.affine, img.affine):
                label_img = resample_to_img(label_img, img, interpolation='nearest')
            labels = np.uint16(label_img.dataobj)
        else:
            labels = np.uint8(data > 0)

        uniq_labels = np.unique(labels[labels > 0])
        label_map = self.inputs.mapping or {label: label for label in uniq_labels}

        rois = {label_map.get(label, label): data[labels == label] for label in label_map}
        with sns.axes_style('whitegrid'):
            fig = sns.histplot(rois, bins=50)
            fig.set_xlabel(self.inputs.xlabel)
        plt.savefig(report_file)
        plt.close()


def get_world_pedir(ornt, pe_direction):
    """Return world direction of phase encoding"""
    # TODO: Move to niworkflows
    axes = (('Right', 'Left'), ('Anterior', 'Posterior'), ('Superior', 'Inferior'))
    ax_idcs = {'i': 0, 'j': 1, 'k': 2}

    if pe_direction is not None:
        axcode = ornt[ax_idcs[pe_direction[0]]]
        inv = pe_direction[1:] == '-'

        for ax in axes:
            for flip in (ax, ax[::-1]):
                if flip[not inv].startswith(axcode):
                    return '-'.join(flip)
    LOGGER.warning(
        'Cannot determine world direction of phase encoding. '
        f'Orientation: {ornt}; PE dir: {pe_direction}'
    )
    return 'Could not be determined - assuming Anterior-Posterior'



"""Script for testing the initial preprocessing steps of T1w and T2w."""


def init_workflow(bids_path, output_path, participant_label, workdir=None):
    """Create the preprocessing workflow."""
    from nipype.pipeline import engine as pe
    from nibabies.workflows.anatomical.preproc import init_anat_average_wf
    from nibabies.workflows.anatomical.registration import init_coregistration_wf
    from nibabies.workflows.anatomical.brain_extraction import (
        init_infant_brain_extraction_wf,
    )
    from nibabies.workflows.anatomical.outputs import init_coreg_report_wf

    wf = pe.Workflow(name="nibabies_anat")
    for subid in participant_label:
        sub_wf = pe.Workflow(name=f"nibabies_anat_{subid}")
        t1w_files = list(
            (bids_path / f"sub-{subid}" / "anat").glob(f"sub-{subid}*_T1w.nii.gz")
        )

        t2w_files = list(
            (bids_path / f"sub-{subid}" / "anat").glob(f"sub-{subid}*_T2w.nii.gz")
        )

        t1w_ref = init_anat_average_wf(
            num_maps=len(t1w_files), name="t1w_ref", omp_nthreads=8
        )
        t2w_ref = init_anat_average_wf(
            num_maps=len(t2w_files), name="t2w_ref", omp_nthreads=8
        )

        t1w_ref.inputs.inputnode.in_files = [str(f) for f in t1w_files]
        t2w_ref.inputs.inputnode.in_files = [str(f) for f in t2w_files]

        be = init_infant_brain_extraction_wf(omp_nthreads=8, age_months=2)
        cr = init_coregistration_wf(omp_nthreads=8, sloppy=True)

        rpt = init_coreg_report_wf(output_dir=str(output_path.absolute()))
        rpt.inputs.inputnode.source_file = [str(f) for f in t1w_files]

        # fmt:off
        sub_wf.connect([
            (t2w_ref, be, [("outputnode.out_file", "inputnode.in_t2w")]),
            (t1w_ref, cr, [("outputnode.out_file", "inputnode.in_t1w")]),
            (be, cr, [
                ("outputnode.t2w_preproc", "inputnode.in_t2w_preproc"),
                ("outputnode.out_mask", "inputnode.in_mask"),
                ("outputnode.out_probmap", "inputnode.in_probmap"),
            ]),
            (cr, rpt, [
                ("outputnode.t1w_preproc", "inputnode.t1w_preproc"),
                ("outputnode.t2w_preproc", "inputnode.t2w_preproc"),
                ("outputnode.t1w_mask", "inputnode.in_mask"),
            ]),
        ])
        # fmt:on
        wf.add_nodes([sub_wf])

    if workdir:
        wf.base_dir = workdir
    return wf


if __name__ == "__main__":
    from pathlib import Path
    import re
    from argparse import ArgumentParser
    from argparse import RawTextHelpFormatter

    parser = ArgumentParser(
        description="sMRIPrep-infants: Structural MRI PREProcessing workflows",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "bids_dir",
        action="store",
        type=Path,
        help="the root folder of a BIDS valid dataset (sub-XXXXX folders should "
        "be found at the top level in this folder).",
    )
    parser.add_argument(
        "output_dir",
        action="store",
        type=Path,
        help="the output path for the outcomes of preprocessing and visual " "reports",
    )
    parser.add_argument(
        "--participant-label",
        "--participant_label",
        action="store",
        nargs="+",
        help="a space delimited list of participant identifiers or a single "
        "identifier (the sub- prefix can be removed)",
    )

    opts = parser.parse_args()

    participant_label = [
        re.sub(r"^sub-", "", p) for p in opts.participant_label
    ]
    init_workflow(
        opts.bids_dir,
        opts.output_dir,
        participant_label,
        workdir=Path.cwd().absolute() / "workdir"
    ).run()



from copy import deepcopy
from pathlib import Path
from unittest.mock import patch

import bids
import nibabel as nb
import numpy as np
import pandas as pd
import pytest
from nipype.pipeline.engine.utils import generate_expanded_graph
from niworkflows.utils.testing import generate_bids_skeleton
from sdcflows.fieldmaps import clear_registry
from sdcflows.utils.wrangler import find_estimators

from ... import config
from ..base import get_estimator, init_nibabies_wf
from ..tests import mock_config

BASE_LAYOUT = {
    '01': {
        'anat': [
            {'run': 1, 'suffix': 'T1w'},
            {'run': 2, 'suffix': 'T1w'},
            {'suffix': 'T2w'},
        ],
        'func': [
            *(
                {
                    'task': 'rest',
                    'run': i,
                    'suffix': suffix,
                    'metadata': {
                        'RepetitionTime': 2.0,
                        'PhaseEncodingDirection': 'j',
                        'TotalReadoutTime': 0.6,
                        'EchoTime': 0.03,
                        'SliceTiming': [0.0, 0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8],
                    },
                }
                for suffix in ('bold', 'sbref')
                for i in range(1, 3)
            ),
            *(
                {
                    'task': 'nback',
                    'echo': i,
                    'suffix': 'bold',
                    'metadata': {
                        'RepetitionTime': 2.0,
                        'PhaseEncodingDirection': 'j',
                        'TotalReadoutTime': 0.6,
                        'EchoTime': 0.015 * i,
                        'SliceTiming': [0.0, 0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8],
                    },
                }
                for i in range(1, 4)
            ),
        ],
        'fmap': [
            {'suffix': 'phasediff', 'metadata': {'EchoTime1': 0.005, 'EchoTime2': 0.007}},
            {'suffix': 'magnitude1', 'metadata': {'EchoTime': 0.005}},
            {
                'suffix': 'epi',
                'direction': 'PA',
                'metadata': {'PhaseEncodingDirection': 'j', 'TotalReadoutTime': 0.6},
            },
            {
                'suffix': 'epi',
                'direction': 'AP',
                'metadata': {'PhaseEncodingDirection': 'j-', 'TotalReadoutTime': 0.6},
            },
        ],
    },
}


T1W_ONLY = BASE_LAYOUT.copy()
T1W_ONLY['01']['anat'] = [{'suffix': 'T1w'}]

T2W_ONLY = T1W_ONLY.copy()
T2W_ONLY['01']['anat'] = [{'suffix': 'T2w'}]


@pytest.fixture(scope='module', autouse=True)
def _quiet_logger():
    import logging

    logger = logging.getLogger('nipype.workflow')
    old_level = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)
    yield
    logger.setLevel(old_level)


@pytest.fixture(autouse=True)
def _reset_sdcflows_registry():
    yield
    clear_registry()


@pytest.fixture(scope='module', params=[BASE_LAYOUT, T1W_ONLY, T2W_ONLY])
def bids_root(tmp_path_factory, request):
    """
    Create a BIDS skeleton for various input types:
    - Full inputs (T1w + T2w)
    - T1w only
    - T2w only
    """
    base = tmp_path_factory.mktemp('base')
    bids_dir = base / 'bids'
    generate_bids_skeleton(bids_dir, request.param)

    # Ensure age information is available
    df = pd.DataFrame({'participant_id': ['sub-01'], 'age_months': ['1']})
    df.to_csv(str(bids_dir / 'participants.tsv'), sep='\t', index=False)

    img = nb.Nifti1Image(np.zeros((10, 10, 10, 10)), np.eye(4))

    for bold_path in bids_dir.glob('sub-01/*/*.nii.gz'):
        img.to_filename(bold_path)

    return bids_dir


def _make_params(
    bold2anat_init: str = 'auto',
    use_bbr: bool | None = None,
    dummy_scans: int | None = None,
    me_output_echos: bool = False,
    medial_surface_nan: bool = False,
    project_goodvoxels: bool = False,
    cifti_output: bool | str = False,
    run_msmsulc: bool = True,
    skull_strip_anat: str = 'auto',
    use_syn_sdc: str | bool = False,
    force_syn: bool = False,
    surface_recon_method: str | None = 'auto',
    ignore: list[str] = None,
    bids_filters: dict = None,
    norm_csf: bool = False,
    multi_step_reg: bool = True,
):
    if ignore is None:
        ignore = []
    if bids_filters is None:
        bids_filters = {}
    return (
        bold2anat_init,
        use_bbr,
        dummy_scans,
        me_output_echos,
        medial_surface_nan,
        project_goodvoxels,
        cifti_output,
        run_msmsulc,
        skull_strip_anat,
        use_syn_sdc,
        force_syn,
        surface_recon_method,
        ignore,
        bids_filters,
        norm_csf,
        multi_step_reg,
    )


@pytest.mark.parametrize('level', ['minimal', 'resampling', 'full'])
@pytest.mark.parametrize('anat_only', [False, True])
@pytest.mark.parametrize(
    (
        'bold2anat_init',
        'use_bbr',
        'dummy_scans',
        'me_output_echos',
        'medial_surface_nan',
        'project_goodvoxels',
        'cifti_output',
        'run_msmsulc',
        'skull_strip_anat',
        'use_syn_sdc',
        'force_syn',
        'surface_recon_method',
        'ignore',
        'bids_filters',
        'norm_csf',
        'multi_step_reg',
    ),
    [
        _make_params(),
        _make_params(bold2anat_init='t1w'),
        _make_params(bold2anat_init='t2w'),
        _make_params(bold2anat_init='header'),
        _make_params(use_bbr=True),
        _make_params(use_bbr=False),
        _make_params(bold2anat_init='header', use_bbr=True),
        # Currently disabled
        # _make_params(bold2anat_init="header", use_bbr=False),
        _make_params(dummy_scans=2),
        _make_params(me_output_echos=True),
        _make_params(medial_surface_nan=True),
        _make_params(cifti_output='91k'),
        _make_params(cifti_output='91k', project_goodvoxels=True),
        _make_params(cifti_output='91k', run_msmsulc=False),
        _make_params(skull_strip_anat='force'),
        _make_params(skull_strip_anat='skip'),
        # _make_params(use_syn_sdc='warn', force_syn=True, ignore=['fieldmaps']),
        _make_params(surface_recon_method=None),
        _make_params(surface_recon_method=None, use_bbr=True),
        _make_params(surface_recon_method=None, use_bbr=False),
        _make_params(surface_recon_method='auto'),
        # _make_params(surface_recon_method='mcribs'),  # Requires precomputed segmentation
        _make_params(surface_recon_method='infantfs'),
        _make_params(surface_recon_method='freesurfer'),
        # Currently unsupported:
        # _make_params(freesurfer=False, bold2anat_init="header"),
        # _make_params(freesurfer=False, bold2anat_init="header", use_bbr=True),
        # _make_params(freesurfer=False, bold2anat_init="header", use_bbr=False),
        # Regression test for gh-3154:
        _make_params(bids_filters={'sbref': {'suffix': 'sbref'}}),
        _make_params(norm_csf=True),
        _make_params(multi_step_reg=True),
        _make_params(multi_step_reg=False),
    ],
)
def test_init_nibabies_wf(
    monkeypatch,
    bids_root: Path,
    tmp_path: Path,
    level: str,
    anat_only: bool,
    bold2anat_init: str,
    use_bbr: bool | None,
    dummy_scans: int | None,
    me_output_echos: bool,
    medial_surface_nan: bool,
    project_goodvoxels: bool,
    cifti_output: bool | str,
    run_msmsulc: bool,
    skull_strip_anat: str,
    use_syn_sdc: str | bool,
    force_syn: bool,
    surface_recon_method: str | None,
    ignore: list[str],
    bids_filters: dict,
    norm_csf: bool,
    multi_step_reg: bool,
):
    monkeypatch.setenv('SUBJECTS_DIR', '/opt/freesurfer/subjects')
    with mock_config(bids_dir=bids_root):
        config.workflow.level = level
        config.workflow.anat_only = anat_only
        config.workflow.bold2anat_init = bold2anat_init
        config.workflow.use_bbr = use_bbr
        config.workflow.dummy_scans = dummy_scans
        config.execution.me_output_echos = me_output_echos
        config.workflow.medial_surface_nan = medial_surface_nan
        config.workflow.project_goodvoxels = project_goodvoxels
        config.workflow.norm_csf = norm_csf
        config.workflow.multi_step_reg = multi_step_reg
        # config.workflow.run_msmsulc = run_msmsulc
        config.workflow.skull_strip_anat = skull_strip_anat
        config.workflow.cifti_output = cifti_output
        config.workflow.surface_recon_method = surface_recon_method
        config.workflow.ignore = ignore
        with patch.dict('nibabies.config.execution.bids_filters', bids_filters):
            wf = init_nibabies_wf(
                [
                    ('01', None),
                ]
            )

    generate_expanded_graph(wf._create_flat_graph())


def test_get_estimator_none(tmp_path):
    bids_dir = tmp_path / 'bids'

    # No IntendedFors/B0Fields
    generate_bids_skeleton(bids_dir, BASE_LAYOUT)
    layout = bids.BIDSLayout(bids_dir)
    bold_files = sorted(
        layout.get(suffix='bold', task='rest', extension='.nii.gz', return_type='file')
    )

    assert get_estimator(layout, bold_files[0]) == ()
    assert get_estimator(layout, bold_files[1]) == ()


def test_get_estimator_b0field_and_intendedfor(tmp_path):
    bids_dir = tmp_path / 'bids'

    # Set B0FieldSource for run 1
    spec = deepcopy(BASE_LAYOUT)
    spec['01']['func'][0]['metadata']['B0FieldSource'] = 'epi'
    spec['01']['fmap'][2]['metadata']['B0FieldIdentifier'] = 'epi'
    spec['01']['fmap'][3]['metadata']['B0FieldIdentifier'] = 'epi'

    # Set IntendedFor for run 2
    spec['01']['fmap'][0]['metadata']['IntendedFor'] = 'func/sub-01_task-rest_run-2_bold.nii.gz'

    generate_bids_skeleton(bids_dir, spec)
    layout = bids.BIDSLayout(bids_dir)
    _ = find_estimators(layout=layout, subject='01')

    bold_files = sorted(
        layout.get(suffix='bold', task='rest', extension='.nii.gz', return_type='file')
    )

    assert get_estimator(layout, bold_files[0]) == ('epi',)
    assert get_estimator(layout, bold_files[1]) == ('auto_00000',)


def test_get_estimator_overlapping_specs(tmp_path):
    bids_dir = tmp_path / 'bids'

    # Set B0FieldSource for both runs
    spec = deepcopy(BASE_LAYOUT)
    spec['01']['func'][0]['metadata']['B0FieldSource'] = 'epi'
    spec['01']['func'][1]['metadata']['B0FieldSource'] = 'epi'
    spec['01']['fmap'][2]['metadata']['B0FieldIdentifier'] = 'epi'
    spec['01']['fmap'][3]['metadata']['B0FieldIdentifier'] = 'epi'

    # Set IntendedFor for both runs
    spec['01']['fmap'][0]['metadata']['IntendedFor'] = [
        'func/sub-01_task-rest_run-1_bold.nii.gz',
        'func/sub-01_task-rest_run-2_bold.nii.gz',
    ]

    generate_bids_skeleton(bids_dir, spec)
    layout = bids.BIDSLayout(bids_dir)
    _ = find_estimators(layout=layout, subject='01')

    bold_files = sorted(
        layout.get(suffix='bold', task='rest', extension='.nii.gz', return_type='file')
    )

    # B0Fields take precedence
    assert get_estimator(layout, bold_files[0]) == ('epi',)
    assert get_estimator(layout, bold_files[1]) == ('epi',)


def test_get_estimator_multiple_b0fields(tmp_path):
    bids_dir = tmp_path / 'bids'

    # Set B0FieldSource for both runs
    spec = deepcopy(BASE_LAYOUT)
    spec['01']['func'][0]['metadata']['B0FieldSource'] = ('epi', 'phasediff')
    spec['01']['func'][1]['metadata']['B0FieldSource'] = 'epi'
    spec['01']['fmap'][0]['metadata']['B0FieldIdentifier'] = 'phasediff'
    spec['01']['fmap'][1]['metadata']['B0FieldIdentifier'] = 'phasediff'
    spec['01']['fmap'][2]['metadata']['B0FieldIdentifier'] = 'epi'
    spec['01']['fmap'][3]['metadata']['B0FieldIdentifier'] = 'epi'

    generate_bids_skeleton(bids_dir, spec)
    layout = bids.BIDSLayout(bids_dir)
    _ = find_estimators(layout=layout, subject='01')

    bold_files = sorted(
        layout.get(suffix='bold', task='rest', extension='.nii.gz', return_type='file')
    )

    # Always get an iterable; don't care if it's a list or tuple
    assert get_estimator(layout, bold_files[0]) == ['epi', 'phasediff']
    assert get_estimator(layout, bold_files[1]) == ('epi',)



import logging
import typing as ty
from pathlib import Path

from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.anat.ants import init_n4_only_wf
from niworkflows.engine import Workflow, tag
from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
from niworkflows.interfaces.header import ValidateImage
from niworkflows.interfaces.nibabel import ApplyMask, Binarize
from niworkflows.interfaces.utility import KeySelect
from niworkflows.utils.connections import pop_file
from smriprep.workflows.anatomical import (
    _is_skull_stripped,
    init_anat_ribbon_wf,
    init_anat_template_wf,
)
from smriprep.workflows.fit.registration import init_register_template_wf
from smriprep.workflows.outputs import (
    init_ds_dseg_wf,
    init_ds_fs_registration_wf,
    init_ds_mask_wf,
    init_ds_surface_metrics_wf,
    init_ds_surfaces_wf,
    init_ds_template_registration_wf,
    init_ds_template_wf,
    init_ds_tpms_wf,
)
from smriprep.workflows.surfaces import (
    init_fsLR_reg_wf,
    init_gifti_morphometrics_wf,
    init_gifti_surfaces_wf,
    init_refinement_wf,
)

from nibabies import config
from nibabies.interfaces import DerivativesDataSink
from nibabies.workflows.anatomical.brain_extraction import init_infant_brain_extraction_wf
from nibabies.workflows.anatomical.outputs import init_anat_reports_wf, init_coreg_report_wf
from nibabies.workflows.anatomical.preproc import (
    init_anat_preproc_wf,
    init_conform_derivative_wf,
    init_csf_norm_wf,
)
from nibabies.workflows.anatomical.registration import (
    init_concat_registrations_wf,
    init_coregistration_wf,
)
from nibabies.workflows.anatomical.segmentation import init_segmentation_wf
from nibabies.workflows.anatomical.surfaces import init_mcribs_dhcp_wf

if ty.TYPE_CHECKING:
    from niworkflows.utils.spaces import Reference, SpatialReferences


LOGGER = logging.getLogger('nipype.workflow')


@tag('anat.fit')
def init_infant_anat_fit_wf(
    *,
    age_months: int,
    t1w: list,
    t2w: list,
    flair: list,
    bids_root: str,
    precomputed: dict,
    longitudinal: bool,
    omp_nthreads: int,
    output_dir: str,
    segmentation_atlases: str | Path | None,
    skull_strip_mode: ty.Literal['auto', 'skip', 'force'],
    skull_strip_template: 'Reference',
    skull_strip_fixed_seed: bool,
    sloppy: bool,
    spaces: 'SpatialReferences',
    recon_method: ty.Literal['freesurfer', 'infantfs', 'mcribs'] | None,
    reference_anat: ty.Literal['T1w', 'T2w'],
    cifti_output: ty.Literal['91k', '170k', False],
    msm_sulc: bool = False,
    name: str = 'infant_anat_fit_wf',
) -> Workflow:
    """
    Stage the anatomical preprocessing steps:
    - T1w reference
    - T2w reference
    - Brain extraction and INU (bias field) correction
    - Brain tissue segmentation
    - Spatial normalization to standard spaces.
    - Surface reconstruction (MCRIBS / infant_recon_all / recon-all)
    """

    workflow = Workflow(name=name)
    num_t1w = len(t1w)
    num_t2w = len(t2w)

    if not num_t1w and not num_t2w:
        raise FileNotFoundError('No anatomical scans provided!')

    anat = reference_anat.lower()

    # Organization
    # ------------
    # This workflow takes the usual (inputnode -> graph -> outputnode) format
    # The graph consists of (input -> compute -> datasink -> buffer) units,
    # and all inputs to outputnode are buffer.
    # If precomputed inputs are found, then these units are replaced with (buffer)
    #     At the time of writing, t1w_mask / t2w_mask are an exception, which takes the form
    #     (t{1,2}w_buffer -> refined_buffer -> datasink -> outputnode)
    # All outputnode components should therefore point to files in the input or
    # output directories.
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['t1w', 't2w', 'roi', 'flair', 'subjects_dir', 'subject_id'],
        ),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                # Primary derivatives
                # T1w
                't1w_preproc',
                't1w_mask',
                't1w_valid_list',
                # T2w
                't2w_preproc',
                't2w_mask',
                't2w_valid_list',
                # Anat specific
                'anat_preproc',
                'anat_mask',
                'anat_dseg',
                'anat_tpms',
                'anat2std_xfm',
                'fsnative2anat_xfm',
                't1w2t2w_xfm',
                't2w2t1w_xfm',
                # Surface and metric derivatives for fsLR resampling
                'white',
                'pial',
                'midthickness',
                'sphere',
                'thickness',
                'sulc',
                'sphere_reg',
                'sphere_reg_fsLR',
                'sphere_reg_msm',
                'anat_ribbon',
                # Reverse transform; not computable from forward transform
                'std2anat_xfm',
                # Metadata
                'template',
                'subjects_dir',
                'subject_id',
                'anat_valid_list',
            ]
        ),
        name='outputnode',
    )

    # If all derivatives exist, inputnode could go unconnected, so add explicitly
    workflow.add_nodes([inputnode])

    # Duplicates will be added to the sourcefile / anat buffers as `anat_...` to facilitate
    # usage of the preferred anatomical

    # Stage 1 inputs (filtered)
    sourcefile_buffer = pe.Node(
        niu.IdentityInterface(
            fields=['t1w_source_files', 't2w_source_files', 'anat_source_files'],
        ),
        name='sourcefile_buffer',
    )

    # Stage 2 - Anatomicals
    t1w_buffer = pe.Node(
        niu.IdentityInterface(fields=['t1w_preproc', 't1w_mask', 't1w_brain']),
        name='t1w_buffer',
    )
    t2w_buffer = pe.Node(
        niu.IdentityInterface(fields=['t2w_preproc', 't2w_mask', 't2w_brain', 't2w_probmap']),
        name='t2w_buffer',
    )
    anat_buffer = pe.Node(
        niu.IdentityInterface(
            fields=[
                'anat_preproc',
                'anat_mask',
                'anat_brain',
            ]
        ),
        name='anat_buffer',
    )

    # Additional buffer if CSF normalization is used
    anat_reg_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_preproc']),
        name='anat_reg_buffer',
    )
    if not config.workflow.norm_csf:
        workflow.connect(anat_buffer, 'anat_preproc', anat_reg_buffer, 'anat_preproc')

    if reference_anat == 'T1w':
        LOGGER.info('ANAT: Using T1w as the reference anatomical')
        workflow.connect([
            (t1w_buffer, anat_buffer, [
                ('t1w_preproc', 'anat_preproc'),
                ('t1w_mask', 'anat_mask'),
                ('t1w_brain', 'anat_brain'),
            ]),
        ])  # fmt:skip
    elif reference_anat == 'T2w':
        LOGGER.info('ANAT: Using T2w as the reference anatomical')
        workflow.connect([
            (t2w_buffer, anat_buffer, [
                ('t2w_preproc', 'anat_preproc'),
                ('t2w_mask', 'anat_mask'),
                ('t2w_brain', 'anat_brain'),
            ]),
        ])  # fmt:skip

    # Stage 3 - Coregistration transforms
    coreg_buffer = pe.Node(
        niu.IdentityInterface(fields=['t1w2t2w_xfm', 't2w2t1w_xfm']),
        name='coreg_buffer',
    )

    aseg_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_aseg']),
        name='aseg_buffer',
    )

    # Stage 4 - Segmentation
    seg_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_dseg', 'anat_tpms', 'ants_segs']),
        name='seg_buffer',
    )
    # Stage 5 - collated template names, forward and reverse transforms
    template_buffer = pe.Node(niu.Merge(3), name='template_buffer')
    anat2std_buffer = pe.Node(niu.Merge(3), name='anat2std_buffer')
    std2anat_buffer = pe.Node(niu.Merge(3), name='std2anat_buffer')

    # Stage 6 results: Refined stage 2 results; may be direct copy if no refinement
    refined_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_mask', 'anat_brain']),
        name='refined_buffer',
    )

    fsnative_buffer = pe.Node(
        niu.IdentityInterface(fields=['fsnative2anat_xfm', 'anat2fsnative_xfm']),
        name='fsnative_buffer',
    )

    # Stage 8 results: GIFTI surfaces
    surfaces_buffer = pe.Node(
        niu.IdentityInterface(
            fields=['white', 'pial', 'midthickness', 'sphere', 'sphere_reg', 'thickness', 'sulc']
        ),
        name='surfaces_buffer',
    )

    # Stage 9 and 10 results: fsLR sphere registration
    fsLR_buffer = pe.Node(niu.IdentityInterface(fields=['sphere_reg_fsLR']), name='fsLR_buffer')
    msm_buffer = pe.Node(niu.IdentityInterface(fields=['sphere_reg_msm']), name='msm_buffer')

    workflow.connect([
        (anat_buffer, outputnode, [
            ('anat_preproc', 'anat_preproc'),
        ]),
        (refined_buffer, outputnode, [
            ('anat_mask', 'anat_mask'),
            ('anat_brain', 'anat_brain'),
        ]),
        (seg_buffer, outputnode, [
            ('anat_dseg', 'anat_dseg'),
            ('anat_tpms', 'anat_tpms'),
        ]),
        (anat2std_buffer, outputnode, [('out', 'anat2std_xfm')]),
        (std2anat_buffer, outputnode, [('out', 'std2anat_xfm')]),
        (template_buffer, outputnode, [('out', 'template')]),
        (sourcefile_buffer, outputnode, [
            ('t1w_source_files', 't1w_valid_list'),
            ('t2w_source_files', 't2w_valid_list'),
            ('anat_source_files', 'anat_valid_list'),  # Alias for reference anat files
        ]),
        (surfaces_buffer, outputnode, [
            ('white', 'white'),
            ('pial', 'pial'),
            ('midthickness', 'midthickness'),
            ('sphere', 'sphere'),
            ('sphere_reg', 'sphere_reg'),
            ('thickness', 'thickness'),
            ('sulc', 'sulc'),
        ]),
        (fsLR_buffer, outputnode, [('sphere_reg_fsLR', 'sphere_reg_fsLR')]),
        (msm_buffer, outputnode, [('sphere_reg_msm', 'sphere_reg_msm')]),
    ])  # fmt:skip

    # Reporting
    anat_reports_wf = init_anat_reports_wf(
        spaces=spaces,
        surface_recon=recon_method,
        output_dir=output_dir,
        sloppy=sloppy,
    )

    workflow.connect([
        (outputnode, anat_reports_wf, [
            ('anat_valid_list', 'inputnode.source_file'),
            ('anat_preproc', 'inputnode.anat_preproc'),
            ('anat_mask', 'inputnode.anat_mask'),
            ('anat_dseg', 'inputnode.anat_dseg'),
            ('template', 'inputnode.template'),
            ('anat2std_xfm', 'inputnode.anat2std_xfm'),
            ('subjects_dir', 'inputnode.subjects_dir'),
            ('subject_id', 'inputnode.subject_id'),
        ]),
    ])  # fmt:skip

    desc = (
        '\nAnatomical data preprocessing\n\n: '
        f'A total of {len(t1w)} T1w and {len(t2w)} T2w images '
        'were found within the input BIDS dataset.'
    )

    t1w_preproc = precomputed.get('t1w_preproc')
    t2w_preproc = precomputed.get('t2w_preproc')
    anat_preproc = precomputed.get(f'{anat}_preproc')

    # Stage 1: Conform & valid T1w/T2w images
    # Note: Since stage 1 & 2 are tightly knit together, it may be more intuitive
    # to combine them in a separate workflow.
    t1w_validate = pe.Node(ValidateImage(), name='t1w_validate', run_without_submitting=True)
    t2w_validate = t1w_validate.clone('t2w_validate')

    if not t1w_preproc:
        LOGGER.info('ANAT Stage 1: Adding T1w template workflow')
        desc += (
            'The T1-weighted (T1w) image was denoised and corrected for intensity '
            'non-uniformity (INU)'
        )

        t1w_template_wf = init_anat_template_wf(
            image_type='T1w',
            num_files=num_t1w,
            longitudinal=longitudinal,
            omp_nthreads=omp_nthreads,
            name='t1w_template_wf',
        )
        ds_t1w_template_wf = init_ds_template_wf(
            image_type='T1w',
            output_dir=output_dir,
            num_anat=num_t1w,
            name='ds_t1w_template_wf',
        )

        if reference_anat == 'T1w':
            workflow.connect([
                (t1w_template_wf, sourcefile_buffer, [
                    ('outputnode.anat_valid_list', 'anat_source_files'),
                ]),
                (t1w_template_wf, anat_reports_wf, [
                    ('outputnode.out_report', 'inputnode.anat_conform_report'),
                ]),
            ])  # fmt:skip

        workflow.connect([
            (inputnode, t1w_template_wf, [('t1w', 'inputnode.anat_files')]),
            (t1w_template_wf, t1w_validate, [('outputnode.anat_ref', 'in_file')]),
            (t1w_template_wf, sourcefile_buffer, [
                ('outputnode.anat_valid_list', 't1w_source_files'),
            ]),
            (t1w_template_wf, ds_t1w_template_wf, [
                ('outputnode.anat_realign_xfm', 'inputnode.anat_ref_xfms'),
            ]),
            (sourcefile_buffer, ds_t1w_template_wf, [
                ('t1w_source_files', 'inputnode.source_files'),
            ]),
            (t1w_buffer, ds_t1w_template_wf, [('t1w_preproc', 'inputnode.anat_preproc')]),
            (ds_t1w_template_wf, outputnode, [('outputnode.anat_preproc', 't1w_preproc')]),
        ])  # fmt:skip
    else:
        LOGGER.info('ANAT Found preprocessed T1w - skipping Stage 1')
        desc += ' A preprocessed T1w image was provided as input.'

        t1w_validate.inputs.in_file = t1w_preproc
        sourcefile_buffer.inputs.t1w_source_files = [t1w_preproc]
        if reference_anat == 'T1w':
            sourcefile_buffer.inputs.anat_source_files = [t1w_preproc]

        workflow.connect([
            (t1w_validate, t1w_buffer, [('out_file', 't1w_preproc')]),
        ])  # fmt:skip

    if not t2w_preproc:
        LOGGER.info('ANAT Stage 1: Adding T2w template workflow')
        desc += (
            'The T2-weighted (T2w) image was denoised and corrected for intensity '
            'non-uniformity (INU)'
        )

        t2w_template_wf = init_anat_template_wf(
            image_type='T2w',
            num_files=num_t1w,
            longitudinal=longitudinal,
            omp_nthreads=omp_nthreads,
            name='t2w_template_wf',
        )
        ds_t2w_template_wf = init_ds_template_wf(
            image_type='T2w',
            output_dir=output_dir,
            num_anat=num_t2w,
            name='ds_t2w_template_wf',
        )

        if reference_anat == 'T2w':
            workflow.connect([
                (t2w_template_wf, sourcefile_buffer, [
                    ('outputnode.anat_valid_list', 'anat_source_files'),
                ]),
                (t2w_template_wf, anat_reports_wf, [
                    ('outputnode.out_report', 'inputnode.anat_conform_report'),
                ]),
            ])  # fmt:skip

        workflow.connect([
            (inputnode, t2w_template_wf, [('t2w', 'inputnode.anat_files')]),
            (t2w_template_wf, t2w_validate, [('outputnode.anat_ref', 'in_file')]),
            (t2w_template_wf, sourcefile_buffer, [
                ('outputnode.anat_valid_list', 't2w_source_files'),
            ]),
            (t2w_template_wf, ds_t2w_template_wf, [
                ('outputnode.anat_realign_xfm', 'inputnode.anat_ref_xfms'),
            ]),
            (sourcefile_buffer, ds_t2w_template_wf, [
                ('t2w_source_files', 'inputnode.source_files'),
            ]),
            (t2w_buffer, ds_t2w_template_wf, [('t2w_preproc', 'inputnode.anat_preproc')]),
            (ds_t2w_template_wf, outputnode, [('outputnode.anat_preproc', 't2w_preproc')]),
        ])  # fmt:skip
    else:
        LOGGER.info('ANAT Found preprocessed T2w - skipping Stage 1')
        desc += ' A preprocessed T2w image was provided as input.'

        t2w_validate.inputs.in_file = t2w_preproc
        sourcefile_buffer.inputs.t2w_source_files = [t2w_preproc]
        if reference_anat == 'T2w':
            sourcefile_buffer.inputs.anat_source_files = [t2w_preproc]

        workflow.connect([
            (t2w_validate, t2w_buffer, [('out_file', 't2w_preproc')]),
        ])  # fmt:skip

    # Stage 2: Use previously computed mask or calculate
    # If we only have one mask (could be either T1w/T2w),
    # just apply transform to get it in the other space
    t1w_mask = precomputed.get('t1w_mask')
    t2w_mask = precomputed.get('t2w_mask')
    anat_mask = precomputed.get(f'{anat}_mask')
    refine_mask = False
    # T1w masking - define preemptively
    apply_t1w_mask = pe.Node(ApplyMask(), name='apply_t1w_mask')
    apply_t2w_mask = apply_t1w_mask.clone(name='apply_t2w_mask')

    # T1w masking logic:
    #
    # PCM = Pre-computed mask
    # PCT = Pre-computed template
    # SS = Skull stripping required
    #
    # PCM, PCT, SS -> Apply PCM to PCT
    # PCM, PCT, !SS -> Apply PCM to PCT
    # PCM, !PCT, SS -> Apply PCM to template, then run INU
    # PCM, !PCT, !SS -> Apply PCM to template, then run INU
    # !PCM, PCT, SS -> Transform T2w brain mask with t2w2t1w_xfm
    # !PCM, PCT, !SS -> Binarize PCT
    # !PCM, !PCT, SS -> Transform T2w brain mask with t2w2t1w_xfm
    # !PCM, !PCT, !SS -> INU correct template

    if not t1w_mask:
        if skull_strip_mode == 'auto':
            run_t1w_skull_strip = not all(_is_skull_stripped(img) for img in t1w)
        else:
            run_t1w_skull_strip = {'force': True, 'skip': False}[skull_strip_mode]

        if not run_t1w_skull_strip:
            desc += (
                'The provided T1w image was previously skull-stripped; '
                'a brain mask was derived from the input image.'
            )

            if not t1w_preproc:
                LOGGER.info('ANAT Stage 2: Skipping skull-strip, INU-correction only')

                t1w_n4_only_wf = init_n4_only_wf(
                    omp_nthreads=omp_nthreads,
                    atropos_use_random_seed=not skull_strip_fixed_seed,
                )
                workflow.connect([
                    (t1w_validate, t1w_n4_only_wf, [('out_file', 'inputnode.in_files')]),
                    (t1w_n4_only_wf, t1w_buffer, [
                        (('outputnode.bias_corrected', pop_file), 't1w_preproc'),
                        ('outputnode.out_mask', 't1w_mask'),
                        (('outputnode.out_file', pop_file), 't1w_brain'),
                    ]),
                ])  # fmt:skip

                if reference_anat == 'T1w':
                    refine_mask = True
                    workflow.connect([
                        (t1w_n4_only_wf, seg_buffer, [
                            ('outputnode.out_segm', 'ants_segs'),
                        ]),
                    ])  # fmt:skip
            else:
                LOGGER.info('ANAT Stage 2: Skipping skull-strip, generating mask from input')
                binarize_t1w = pe.Node(Binarize(thresh_low=2), name='binarize_t1w')
                workflow.connect([
                    (t1w_validate, binarize_t1w, [('out_file', 'in_file')]),
                    (t1w_validate, t1w_buffer, [('out_file', 't1w_brain')]),
                    (binarize_t1w, t1w_buffer, [('out_file', 't1w_mask')]),
                ])  # fmt:skip
        else:
            # T2w -> T1w transformation of the mask will occur if either
            # 1) reusing a precomputed T2w mask
            # 2) calculating with T2w template brain extraction
            LOGGER.info('ANAT T2w mask will be transformed into T1w space')
            transform_t2w_mask = pe.Node(
                ApplyTransforms(interpolation='MultiLabel'), name='transform_t2w_mask'
            )
            # TODO: May need to differentiate presence of t1w_preproc
            workflow.connect([
                (t2w_buffer, transform_t2w_mask, [('t2w_mask', 'input_image')]),
                (coreg_buffer, transform_t2w_mask, [('t2w2t1w_xfm', 'transforms')]),
                (t1w_validate, transform_t2w_mask, [('out_file', 'reference_image')]),

                (transform_t2w_mask, t1w_buffer, [('output_image', 't1w_mask')]),
                (transform_t2w_mask, apply_t1w_mask, [('output_image', 'in_mask')]),
                (t1w_validate, apply_t1w_mask, [('out_file', 'in_file')]),
                (apply_t1w_mask, t1w_buffer, [('out_file', 't1w_brain')]),
            ])  # fmt:skip

            if not t1w_preproc:
                t1w_preproc_wf = init_anat_preproc_wf(name='t1w_preproc_wf')
                workflow.connect([
                    (t1w_validate, t1w_preproc_wf, [
                        ('out_file', 'inputnode.in_anat'),
                    ]),
                    (t1w_preproc_wf, t1w_buffer, [
                        ('outputnode.anat_preproc', 't1w_preproc'),
                    ]),
                ])  # fmt:skip

        # Save T1w mask
        ds_t1w_mask_wf = init_ds_mask_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            mask_type='brain',
            extra_entities={'space': 'T1w'},
            name='ds_t1w_mask_wf',
        )
        workflow.connect([
            (sourcefile_buffer, ds_t1w_mask_wf, [('t1w_source_files', 'inputnode.source_files')]),
        ])  # fmt:skip

        if reference_anat == 'T1w':
            workflow.connect([
                (refined_buffer, ds_t1w_mask_wf, [('anat_mask', 'inputnode.mask_file')]),
            ])  # fmt:skip
        else:
            workflow.connect([
                (t1w_buffer, ds_t1w_mask_wf, [('t1w_mask', 'inputnode.mask_file')]),
            ])  # fmt:skip
    else:  # T1w mask derivative has been found
        LOGGER.info('ANAT Found T1w brain mask')
        if reference_anat == 'T1w':
            desc += (
                'A pre-computed T1w brain mask was provided as input and used throughout the '
                'workflow.'
            )
        workflow.connect(apply_t1w_mask, 'out_file', t1w_buffer, 't1w_brain')

        if not t1w_preproc:
            # Ensure compatibility with T1w template
            conform_t1w_mask_wf = init_conform_derivative_wf(
                in_file=t1w_mask, name='conform_t1w_mask_wf'
            )

            LOGGER.info('ANAT Skipping skull-strip, INU-correction only')
            t1w_n4_wf = init_anat_preproc_wf(name='t1w_n4_wf')
            workflow.connect([
                (t1w_validate, conform_t1w_mask_wf, [('out_file', 'inputnode.ref_file')]),
                (conform_t1w_mask_wf, t1w_buffer, [('outputnode.out_file', 't1w_mask')]),
                (conform_t1w_mask_wf, apply_t1w_mask, [('outputnode.out_file', 'in_mask')]),
                (t1w_validate, t1w_n4_wf, [('out_file', 'inputnode.in_anat')]),
                (t1w_n4_wf, t1w_buffer, [('outputnode.anat_preproc', 't1w_preproc')]),
                (t1w_n4_wf, apply_t1w_mask, [('outputnode.anat_preproc', 'in_file')]),
            ])  # fmt:skip
        else:
            LOGGER.info('ANAT Skipping T1w masking')
            workflow.connect(t1w_validate, 'out_file', apply_t1w_mask, 'in_file')
            t1w_buffer.inputs.t1w_mask = t1w_mask
            apply_t1w_mask.inputs.in_mask = t1w_mask

    # T2w masking logic:
    #
    # PCM-T2 = Pre-computed mask
    # PCM-T1 = Pre-computed mask for PCT-T1
    # PCT = Pre-computed template
    # SS = Skull stripping required
    #
    # PCM, PCT, SS -> Apply PCM to PCT
    # PCM, PCT, !SS -> Apply PCM to PCT
    # PCM, !PCT, SS -> Apply PCM to template, then run INU
    # PCM, !PCT, !SS -> Apply PCM to template, then run INU
    # !PCM, PCT, SS -> Run brain extraction
    # !PCM, PCT, !SS -> Binarize PCT
    # !PCM, !PCT, SS -> Run brain extraction
    # !PCM, !PCT, !SS -> INU correct template
    if not t2w_mask:
        if skull_strip_mode == 'auto':
            run_t2w_skull_strip = not all(_is_skull_stripped(img) for img in t2w)
        else:
            run_t2w_skull_strip = {'force': True, 'skip': False}[skull_strip_mode]

        if not run_t2w_skull_strip:
            desc += (
                'The T2w reference was previously skull-stripped; '
                'a brain mask was derived from the input image.'
            )

            if not t2w_preproc:
                LOGGER.info('ANAT Stage 2b: Skipping skull-strip, INU-correction only')

                t2w_n4_only_wf = init_n4_only_wf(
                    omp_nthreads=omp_nthreads,
                    bids_suffix=reference_anat,
                    atropos_use_random_seed=not skull_strip_fixed_seed,
                    name='t2w_n4_only_wf',
                )
                workflow.connect([
                    (t2w_validate, t2w_n4_only_wf, [('out_file', 'inputnode.in_files')]),
                    (t2w_n4_only_wf, t2w_buffer, [
                        (('outputnode.bias_corrected', pop_file), 't2w_preproc'),
                        ('outputnode.out_mask', 't2w_mask'),
                        (('outputnode.out_file', pop_file), 't2w_brain'),
                    ]),
                ])  # fmt:skip
                if reference_anat == 'T2w':
                    refine_mask = True
                    workflow.connect([
                        (t2w_n4_only_wf, seg_buffer, [
                            ('outputnode.out_segm', 'ants_segs'),
                        ]),
                    ])  # fmt:skip
            else:
                LOGGER.info('ANAT Stage 2b: Skipping skull-strip, generating mask from input')
                binarize_t2w = pe.Node(Binarize(thresh_low=2), name='binarize_t2w')
                workflow.connect([
                    (t2w_validate, binarize_t2w, [('out_file', 'in_file')]),
                    (t2w_validate, t2w_buffer, [('out_file', 't2w_brain')]),
                    (binarize_t2w, t2w_buffer, [('out_file', 't2w_mask')]),
                ])  # fmt:skip
        else:
            LOGGER.info('ANAT Atlas-based brain mask will be calculated on the T2w')
            brain_extraction_wf = init_infant_brain_extraction_wf(
                omp_nthreads=omp_nthreads,
                sloppy=sloppy,
                age_months=age_months,
                ants_affine_init=True,
                skull_strip_template=skull_strip_template.space,
                template_specs=skull_strip_template.spec,
                debug='registration' in config.execution.debug,
            )

            workflow.connect([
                (t2w_validate, brain_extraction_wf, [
                    ('out_file', 'inputnode.t2w_preproc'),
                ]),
                (brain_extraction_wf, t2w_buffer, [
                    ('outputnode.out_mask', 't2w_mask'),
                    ('outputnode.t2w_brain', 't2w_brain'),
                    ('outputnode.t2w_preproc', 't2w_preproc'),
                    ('outputnode.out_probmap', 't2w_probmap')
                ]),
            ])  # fmt:skip

        # Save T2w mask
        ds_t2w_mask_wf = init_ds_mask_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            mask_type='brain',
            extra_entities={'space': 'T2w'},
            name='ds_t2w_mask_wf',
        )
        workflow.connect([
            (sourcefile_buffer, ds_t2w_mask_wf, [('t2w_source_files', 'inputnode.source_files')]),
        ])  # fmt:skip

        if reference_anat == 'T2w':
            workflow.connect([
                (refined_buffer, ds_t2w_mask_wf, [('anat_mask', 'inputnode.mask_file')]),
            ])  # fmt:skip
        else:
            workflow.connect([
                (t2w_buffer, ds_t2w_mask_wf, [('t2w_mask', 'inputnode.mask_file')]),
            ])  # fmt:skip
    else:
        LOGGER.info('ANAT Found T2w brain mask')
        if reference_anat == 'T2w':
            desc += (
                'A pre-computed T2w brain mask was provided as input and used throughout the '
                'workflow.'
            )
        workflow.connect(apply_t2w_mask, 'out_file', t2w_buffer, 't2w_brain')

        if not t2w_preproc:
            # Ensure compatibility with T2w template
            conform_t2w_mask_wf = init_conform_derivative_wf(
                in_file=t2w_mask,
                name='conform_t2w_mask_wf',
            )

            LOGGER.info('ANAT Skipping skull-strip, INU-correction only')
            t2w_n4_wf = init_anat_preproc_wf(name='t2w_n4_wf')
            workflow.connect([
                (t2w_validate, conform_t2w_mask_wf, [('out_file', 'inputnode.ref_file')]),
                (conform_t2w_mask_wf, t2w_buffer, [('outputnode.out_file', 't2w_mask')]),
                (conform_t2w_mask_wf, apply_t2w_mask, [('outputnode.out_file', 'in_mask')]),
                (t2w_validate, t2w_n4_wf, [('out_file', 'inputnode.in_anat')]),
                (t2w_n4_wf, t2w_buffer, [('outputnode.anat_preproc', 't2w_preproc')]),
                (t2w_n4_wf, apply_t2w_mask, [('outputnode.anat_preproc', 'in_file')]),
            ])  # fmt:skip
        else:
            LOGGER.info('ANAT Skipping T2w masking')
            workflow.connect(t2w_validate, 'out_file', apply_t2w_mask, 'in_file')
            t2w_buffer.inputs.t2w_mask = t2w_mask
            apply_t2w_mask.inputs.in_mask = t2w_mask

    # Stage 3: Coregistration
    t1w2t2w_xfm = precomputed.get('t1w2t2w_xfm')
    t2w2t1w_xfm = precomputed.get('t2w2t1w_xfm')

    # To use the found xfm, requires both precomputed anatomicals to be found as well
    if (t1w_preproc and t2w_preproc) and (t1w2t2w_xfm and t2w2t1w_xfm):
        LOGGER.info('ANAT Found T1w<->T2w xfms')
        desc += (
            ' A T1w-T2w coregistration transform was provided as input and used throughout '
            'the workflow.'
        )
        coreg_buffer.inputs.t1w2t2w_xfm = t1w2t2w_xfm
        coreg_buffer.inputs.t2w2t1w_xfm = t2w2t1w_xfm
    else:
        LOGGER.info('ANAT Coregistering anatomicals')
        desc += ' The T1w and T2w reference volumes were co-registered using ANTs.'

        probmap = not t2w_preproc and not t2w_mask
        coregistration_wf = init_coregistration_wf(
            omp_nthreads=omp_nthreads,
            sloppy=sloppy,
            debug='registration' in config.execution.debug,
            t1w_mask=False,
            probmap=probmap,
        )

        ds_t1w2t2w_xfm = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                to='T2w',
                mode='image',
                suffix='xfm',
                dismiss_entities=('desc', 'echo'),
                **{'from': 'T1w'},
            ),
            name='ds_t1w2t2w_xfm',
            run_without_submitting=True,
        )

        ds_t2w2t1w_xfm = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                to='T1w',
                mode='image',
                suffix='xfm',
                dismiss_entities=('desc', 'echo'),
                **{'from': 'T2w'},
            ),
            name='ds_t2w2t1w_xfm',
            run_without_submitting=True,
        )

        coreg_report_wf = init_coreg_report_wf(output_dir=output_dir)

        workflow.connect([
            (t1w_validate, coregistration_wf, [
                ('out_file', 'inputnode.in_t1w'),
            ]),
            (t2w_buffer, coregistration_wf, [
                ('t2w_preproc', 'inputnode.in_t2w'),
                ('t2w_mask', 'inputnode.in_mask'),
            ]),
            (coregistration_wf, ds_t1w2t2w_xfm, [
                ('outputnode.t1w2t2w_xfm', 'in_file'),
            ]),
            (sourcefile_buffer, ds_t1w2t2w_xfm, [
                ('t1w_source_files', 'source_file'),
            ]),
            (coregistration_wf, ds_t2w2t1w_xfm, [
                ('outputnode.t2w2t1w_xfm', 'in_file'),
            ]),
            (sourcefile_buffer, ds_t2w2t1w_xfm, [
                ('t2w_source_files', 'source_file'),
            ]),
            (coregistration_wf, coreg_buffer, [
                ('outputnode.t1w2t2w_xfm', 't1w2t2w_xfm'),
                ('outputnode.t2w2t1w_xfm', 't2w2t1w_xfm'),
            ]),
            (coregistration_wf, coreg_report_wf, [
                ('outputnode.t1w_preproc', 'inputnode.t1w_preproc'),
                ('outputnode.t2w_preproc', 'inputnode.t2w_preproc'),
                ('outputnode.t1w_mask', 'inputnode.in_mask'),
            ]),
            (sourcefile_buffer, coreg_report_wf, [('anat_source_files', 'inputnode.source_file')]),
        ])  # fmt:skip

        if probmap:
            workflow.connect([
                (t2w_buffer, coregistration_wf, [
                    ('t2w_probmap', 'inputnode.in_probmap'),
                ])
            ])  # fmt:skip

    # Stage 4: Segmentation
    anat_dseg = precomputed.get(f'{anat}_dseg')
    anat_tpms = precomputed.get(f'{anat}_tpms')
    anat_aseg = precomputed.get(f'{anat}_aseg')

    seg_method = 'jlf' if config.execution.segmentation_atlases_dir else 'fast'

    if anat_aseg:
        LOGGER.info('ANAT Found precomputed anatomical segmentation')
        # Ensure compatibility with anatomical template
        if not anat_preproc:
            conform_aseg_wf = init_conform_derivative_wf(
                in_file=anat_aseg,
                name='conform_aseg_wf',
            )

            workflow.connect([
                (anat_buffer, conform_aseg_wf, [('anat_preproc', 'inputnode.ref_file')]),
                (conform_aseg_wf, aseg_buffer, [('outputnode.out_file', 'anat_aseg')]),
            ])  # fmt:skip
        else:
            aseg_buffer.inputs.anat_aseg = anat_aseg

    if not (anat_dseg and anat_tpms):
        LOGGER.info('ANAT Stage 4: Tissue segmentation')
        segmentation_wf = init_segmentation_wf(
            sloppy=sloppy,
            method=seg_method,
            image_type=reference_anat,
            omp_nthreads=omp_nthreads,
            has_aseg=bool(anat_aseg),
        )

        workflow.connect([
            (anat_buffer, segmentation_wf, [('anat_brain', 'inputnode.anat_brain')]),
            (segmentation_wf, seg_buffer, [
                ('outputnode.anat_dseg', 'anat_dseg'),
                ('outputnode.anat_tpms', 'anat_tpms'),
            ]),
        ])  # fmt:skip

        if anat_aseg:
            workflow.connect(aseg_buffer, 'anat_aseg', segmentation_wf, 'inputnode.anat_aseg')
        elif seg_method == 'jlf':
            workflow.connect(segmentation_wf, 'outputnode.anat_aseg', aseg_buffer, 'anat_aseg')
            # TODO: datasink aseg

        if not anat_dseg:
            ds_dseg_wf = init_ds_dseg_wf(
                output_dir=str(output_dir),
                extra_entities={'space': reference_anat},
            )
            workflow.connect([
                (sourcefile_buffer, ds_dseg_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (segmentation_wf, ds_dseg_wf, [
                    ('outputnode.anat_dseg', 'inputnode.anat_dseg'),
                ]),
            ])  # fmt:skip

        if not anat_tpms:
            ds_tpms_wf = init_ds_tpms_wf(
                output_dir=str(output_dir),
                extra_entities={'space': reference_anat},
            )
            workflow.connect([
                (sourcefile_buffer, ds_tpms_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (segmentation_wf, ds_tpms_wf, [
                    ('outputnode.anat_tpms', 'inputnode.anat_tpms'),
                ]),
            ])  # fmt:skip
    else:
        LOGGER.info('ANAT Stage 4: Skipping segmentation workflow')
    if anat_dseg:
        LOGGER.info('ANAT Found discrete segmentation')
        desc += 'Precomputed discrete tissue segmentations were provided as inputs.\n'
        seg_buffer.inputs.anat_dseg = anat_dseg
    if anat_tpms:
        LOGGER.info('ANAT Found tissue probability maps')
        desc += 'Precomputed tissue probabiilty maps were provided as inputs.\n'
        seg_buffer.inputs.anat_tpms = anat_tpms

    # Stage 5: Normalization

    # If selected adult templates are requested (MNI152 6th Gen or 2009)
    # opt to concatenate transforms first from native -> infant template (MNIInfant),
    # and then use a previously computed MNIInfant<cohort> -> MNI transform
    # this minimizes the chance of a bad registration.

    templates = []
    concat_xfms = []
    found_xfms = {}
    intermediate = None  # The intermediate space when concatenating xfms - includes cohort
    intermediate_targets = (
        {
            'MNI152NLin6Asym',  # TODO: 'MNI152NLin2009cAsym'
        }
        if config.workflow.multi_step_reg
        else set()
    )

    for template in spaces.get_spaces(nonstandard=False, dim=(3,)):
        # resolution / spec will not differentiate here
        if template.startswith('MNIInfant'):
            intermediate = template
        xfms = precomputed.get('transforms', {}).get(template, {})
        if set(xfms) != {'forward', 'reverse'}:
            if template in intermediate_targets:
                concat_xfms.append(template)
            else:
                templates.append(template)
        else:
            found_xfms[template] = xfms

    # Create another set of buffers to handle the case where we aggregate found and generated
    # xfms to be concatenated
    concat_template_buffer = pe.Node(niu.Merge(2), name='concat_template_buffer')
    concat_template_buffer.inputs.in1 = list(found_xfms)
    concat_anat2std_buffer = pe.Node(niu.Merge(2), name='concat_anat2std_buffer')
    concat_anat2std_buffer.inputs.in1 = [xfm['forward'] for xfm in found_xfms.values()]
    concat_std2anat_buffer = pe.Node(niu.Merge(2), name='concat_std2anat_buffer')
    concat_std2anat_buffer.inputs.in1 = [xfm['reverse'] for xfm in found_xfms.values()]

    template_buffer.inputs.in1 = list(found_xfms)
    anat2std_buffer.inputs.in1 = [xfm['forward'] for xfm in found_xfms.values()]
    std2anat_buffer.inputs.in1 = [xfm['reverse'] for xfm in found_xfms.values()]

    if config.workflow.norm_csf:
        csf_norm_wf = init_csf_norm_wf()

        workflow.connect([
            (anat_buffer, csf_norm_wf, [('anat_preproc', 'inputnode.anat_preproc')]),
            (seg_buffer, csf_norm_wf, [('anat_tpms', 'inputnode.anat_tpms')]),
            (csf_norm_wf, anat_reg_buffer, [('outputnode.anat_preproc', 'anat_preproc')]),
        ])  # fmt:skip

    if templates:
        LOGGER.info(f'ANAT Stage 5a: Preparing normalization workflow for {templates}')
        register_template_wf = init_register_template_wf(
            sloppy=sloppy,
            omp_nthreads=omp_nthreads,
            templates=templates,
            image_type=reference_anat,
        )
        ds_template_registration_wf = init_ds_template_registration_wf(
            output_dir=str(output_dir),
            image_type=reference_anat,
        )

        workflow.connect([
            (inputnode, register_template_wf, [('roi', 'inputnode.lesion_mask')]),
            (anat_reg_buffer, register_template_wf, [
                ('anat_preproc', 'inputnode.moving_image'),
            ]),
            (refined_buffer, register_template_wf, [('anat_mask', 'inputnode.moving_mask')]),
            (sourcefile_buffer, ds_template_registration_wf, [
                ('anat_source_files', 'inputnode.source_files')
            ]),
            (register_template_wf, ds_template_registration_wf, [
                ('outputnode.template', 'inputnode.template'),
                ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('outputnode.std2anat_xfm', 'inputnode.std2anat_xfm'),
            ]),
            (register_template_wf, template_buffer, [('outputnode.template', 'in2')]),
            (register_template_wf, concat_template_buffer, [('outputnode.template', 'in2')]),
            (register_template_wf, std2anat_buffer, [('outputnode.std2anat_xfm', 'in2')]),
            (register_template_wf, concat_std2anat_buffer, [('outputnode.std2anat_xfm', 'in2')]),
            (register_template_wf, anat2std_buffer, [('outputnode.anat2std_xfm', 'in2')]),
            (register_template_wf, concat_anat2std_buffer, [('outputnode.anat2std_xfm', 'in2')]),
        ])  # fmt:skip

    if concat_xfms:
        LOGGER.info(f'ANAT Stage 5b: Concatenating normalization for {concat_xfms}')

        select_infant_mni = pe.Node(
            KeySelect(fields=['anat2std_xfm', 'std2anat_xfm'], key=intermediate),
            name='select_infant_mni',
            run_without_submitting=True,
        )
        concat_reg_wf = init_concat_registrations_wf(templates=concat_xfms)
        ds_concat_reg_wf = init_ds_template_registration_wf(
            output_dir=str(output_dir),
            image_type=reference_anat,
            name='ds_concat_registration_wf',
        )

        workflow.connect([
            (concat_template_buffer, select_infant_mni, [('out', 'keys')]),
            (concat_anat2std_buffer, select_infant_mni, [('out', 'anat2std_xfm')]),
            (concat_std2anat_buffer, select_infant_mni, [('out', 'std2anat_xfm')]),
            (select_infant_mni, concat_reg_wf, [
                ('key', 'inputnode.intermediate'),
                ('anat2std_xfm', 'inputnode.anat2int_xfm'),
                ('std2anat_xfm', 'inputnode.int2anat_xfm'),
            ]),
            (sourcefile_buffer, ds_concat_reg_wf, [
                ('anat_source_files', 'inputnode.source_files')
            ]),
            (concat_reg_wf, ds_concat_reg_wf, [
                ('outputnode.template', 'inputnode.template'),
                ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('outputnode.std2anat_xfm', 'inputnode.std2anat_xfm'),
            ]),
            (concat_reg_wf, template_buffer, [('outputnode.template', 'in3')]),
            (concat_reg_wf, anat2std_buffer, [('outputnode.anat2std_xfm', 'in3')]),
            (concat_reg_wf, std2anat_buffer, [('outputnode.std2anat_xfm', 'in3')]),
        ])  # fmt:skip

    if found_xfms:
        LOGGER.info(f'ANAT Stage 5c: Found pre-computed registrations for {found_xfms}')

    # Only refine mask if necessary
    if anat_mask or recon_method is None or not refine_mask:
        workflow.connect([
            (anat_buffer, refined_buffer, [
                ('anat_mask', 'anat_mask'),
                ('anat_brain', 'anat_brain'),
            ]),
        ])  # fmt:skip

    workflow.__desc__ = desc

    if recon_method is None:
        LOGGER.info('ANAT Skipping Stages 6+')
        return workflow

    # Stage 6: Surface reconstruction
    if recon_method == 'mcribs':
        from nibabies.workflows.anatomical.surfaces import init_mcribs_surface_recon_wf

        LOGGER.info('ANAT Stage 6: Preparing M-CRIB-S reconstruction workflow')
        surface_recon_wf = init_mcribs_surface_recon_wf(
            omp_nthreads=omp_nthreads,
            use_aseg=bool(anat_aseg),
            precomputed=precomputed,
            mcribs_dir=str(config.execution.mcribs_dir),
        )

        workflow.connect([
            (inputnode, surface_recon_wf, [
                ('subject_id', 'inputnode.subject_id'),
                ('subjects_dir', 'inputnode.subjects_dir'),
            ]),
            (t2w_validate, surface_recon_wf, [('out_file', 'inputnode.t2w')]),
            (t2w_buffer, surface_recon_wf, [('t2w_mask', 'inputnode.in_mask'),]),
            (aseg_buffer, surface_recon_wf, [
                ('anat_aseg', 'inputnode.in_aseg'),
            ]),
            (surface_recon_wf, outputnode, [
                ('outputnode.subjects_dir', 'subjects_dir'),
                ('outputnode.subject_id', 'subject_id'),
            ]),
        ])  # fmt:skip

    else:
        from smriprep.utils.misc import fs_isRunning

        fs_isrunning = pe.Node(
            niu.Function(function=fs_isRunning), overwrite=True, name='fs_isrunning'
        )
        fs_isrunning.inputs.logger = LOGGER

        if recon_method == 'freesurfer':
            from smriprep.workflows.surfaces import init_surface_recon_wf

            LOGGER.info('ANAT Stage 6: Preparing FreeSurfer recon-all workflow')
            fs_isrunning = pe.Node(
                niu.Function(function=fs_isRunning), overwrite=True, name='fs_isrunning'
            )
            fs_isrunning.inputs.logger = LOGGER

            surface_recon_wf = init_surface_recon_wf(
                name='surface_recon_wf',
                omp_nthreads=omp_nthreads,
                hires=True,
                fs_no_resume=False,
                precomputed=precomputed,
            )

            if t2w or flair:
                t2w_or_flair = 'T2-weighted' if t2w else 'FLAIR'
                if surface_recon_wf.__desc__:
                    surface_recon_wf.__desc__ += (
                        f'A {t2w_or_flair} image was used to improve pial surface refinement.'
                    )
            workflow.connect([
                (inputnode, surface_recon_wf, [
                    ('t2w', 'inputnode.t2w'),
                    ('flair', 'inputnode.flair'),
                ]),
            ])  # fmt:skip

        elif recon_method == 'infantfs':
            from nibabies.workflows.anatomical.surfaces import init_infantfs_surface_recon_wf

            LOGGER.info('ANAT Stage 6: Preparing Infant FreeSurfer workflow')
            surface_recon_wf = init_infantfs_surface_recon_wf(
                age_months=age_months,
                precomputed=precomputed,
                omp_nthreads=omp_nthreads,
                use_aseg=bool(anat_aseg),
            )

        # Force use of the T1w image
        workflow.connect([
            (inputnode, fs_isrunning, [
                ('subjects_dir', 'subjects_dir'),
                ('subject_id', 'subject_id'),
            ]),
            (inputnode, surface_recon_wf, [
                ('subject_id', 'inputnode.subject_id'),
            ]),
            (fs_isrunning, surface_recon_wf, [('out', 'inputnode.subjects_dir')]),
            (t1w_validate, surface_recon_wf, [('out_file', 'inputnode.t1w')]),
            (t1w_buffer, surface_recon_wf, [('t1w_brain', 'inputnode.skullstripped_t1')]),
            (surface_recon_wf, outputnode, [
                ('outputnode.subjects_dir', 'subjects_dir'),
                ('outputnode.subject_id', 'subject_id'),
            ]),
        ])  # fmt:skip

        if anat_aseg:
            workflow.connect(aseg_buffer, 'anat_aseg', surface_recon_wf, 'inputnode.in_aseg')

    fsnative_xfms = precomputed.get('transforms', {}).get('fsnative')
    if not fsnative_xfms:
        ds_fs_registration_wf = init_ds_fs_registration_wf(
            image_type=reference_anat, output_dir=output_dir
        )

        if recon_method == 'freesurfer':
            workflow.connect([
                (surface_recon_wf, fsnative_buffer, [
                    ('outputnode.fsnative2t1w_xfm', 'fsnative2anat_xfm'),
                    ('outputnode.t1w2fsnative_xfm', 'anat2fsnative_xfm'),
                ]),
            ])  # fmt:skip
        else:
            workflow.connect([
                (surface_recon_wf, fsnative_buffer, [
                    ('outputnode.fsnative2anat_xfm', 'fsnative2anat_xfm'),
                    ('outputnode.anat2fsnative_xfm', 'anat2fsnative_xfm'),
                ]),
            ])  # fmt:skip

        workflow.connect([
            (sourcefile_buffer, ds_fs_registration_wf, [
                ('anat_source_files', 'inputnode.source_files'),
            ]),
            (fsnative_buffer, ds_fs_registration_wf, [
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (fsnative_buffer, outputnode, [
                ('fsnative2anat_xfm', 'fsnative2anat_xfm'),
            ]),
        ])  # fmt:skip
    elif 'reverse' in fsnative_xfms:
        LOGGER.info('ANAT Found fsnative-to-anatomical transform - skipping registration')
        outputnode.inputs.fsnative2anat_xfm = fsnative_xfms['reverse']
    else:
        raise RuntimeError(
            'Found an anatomical-to-fsnative transform without the reverse. Time to handle this.'
        )

    if not anat_mask and refine_mask:
        LOGGER.info('ANAT Stage 7: Preparing mask refinement workflow')
        # Stage 6: Refine ANTs mask with FreeSurfer segmentation
        refinement_wf = init_refinement_wf()
        applyrefined = pe.Node(ApplyMask(), name='applyrefined')

        workflow.connect([
            (surface_recon_wf, refinement_wf, [
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
                ('outputnode.subject_id', 'inputnode.subject_id'),
            ]),
            (fsnative_buffer, refinement_wf, [
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (anat_buffer, refinement_wf, [
                ('anat_preproc', 'inputnode.reference_image'),
            ]),
            (seg_buffer, refinement_wf, [
                ('ants_segs', 'inputnode.ants_segs'),  # TODO: Verify this is the same as dseg
            ]),
            (anat_buffer, applyrefined, [('anat_preproc', 'in_file')]),
            (refinement_wf, applyrefined, [('outputnode.out_brainmask', 'in_mask')]),
            (refinement_wf, refined_buffer, [('outputnode.out_brainmask', 'anat_mask')]),
            (applyrefined, refined_buffer, [('out_file', 'anat_brain')]),
        ])  # fmt:skip
    elif not refine_mask:
        LOGGER.info('ANAT Skipping mask refinement workflow')
    else:
        LOGGER.info('ANAT Found brain mask - skipping Stage 7')

    # Stages 8-10: Surface conversion and registration
    # sphere_reg is needed to generate sphere_reg_fsLR
    # sphere and sulc are needed to generate sphere_reg_msm
    # white, pial, midthickness and thickness are needed to resample in the cortical ribbon
    # TODO: Consider paring down or splitting into a subworkflow that can be called on-demand
    # A subworkflow would still need to check for precomputed outputs
    needed_anat_surfs = ['white', 'pial', 'midthickness']
    needed_metrics = ['thickness', 'sulc']
    needed_spheres = ['sphere_reg', 'sphere']

    # Detect pre-computed surfaces
    found_surfs = {
        surf: sorted(precomputed[surf])
        for surf in needed_anat_surfs + needed_metrics + needed_spheres
        if len(precomputed.get(surf, [])) == 2
    }
    if found_surfs:
        LOGGER.info(f'ANAT Stage 8: Found pre-converted surfaces for {list(found_surfs)}')
        surfaces_buffer.inputs.trait_set(**found_surfs)

    # Stage 8: Surface conversion
    surfs = [surf for surf in needed_anat_surfs if surf not in found_surfs]
    spheres = [sphere for sphere in needed_spheres if sphere not in found_surfs]
    if surfs or spheres:
        LOGGER.info(f'ANAT Stage 8: Creating GIFTI surfaces for {surfs + spheres}')
    if surfs:
        gifti_surfaces_wf = init_gifti_surfaces_wf(surfaces=surfs)
        ds_surfaces_wf = init_ds_surfaces_wf(output_dir=output_dir, surfaces=surfs)

        workflow.connect([
            (surface_recon_wf, gifti_surfaces_wf, [
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
            ]),
            (fsnative_buffer, gifti_surfaces_wf, [
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (gifti_surfaces_wf, surfaces_buffer, [
                (f'outputnode.{surf}', surf) for surf in surfs
            ]),
            (sourcefile_buffer, ds_surfaces_wf, [('anat_source_files', 'inputnode.source_files')]),
            (gifti_surfaces_wf, ds_surfaces_wf, [
                (f'outputnode.{surf}', f'inputnode.{surf}') for surf in surfs
            ]),
        ])  # fmt:skip
    if spheres:
        gifti_spheres_wf = init_gifti_surfaces_wf(
            surfaces=spheres, to_scanner=False, name='gifti_spheres_wf'
        )
        ds_spheres_wf = init_ds_surfaces_wf(
            output_dir=output_dir,
            surfaces=spheres,
            name='ds_spheres_wf',
        )

        workflow.connect([
            (surface_recon_wf, gifti_spheres_wf, [
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
                # No transform for spheres, following HCP pipelines' lead
            ]),
            (gifti_spheres_wf, surfaces_buffer, [
                (f'outputnode.{sphere}', sphere) for sphere in spheres
            ]),
            (sourcefile_buffer, ds_spheres_wf, [('anat_source_files', 'inputnode.source_files')]),
            (gifti_spheres_wf, ds_spheres_wf, [
                (f'outputnode.{sphere}', f'inputnode.{sphere}') for sphere in spheres
            ]),
        ])  # fmt:skip
    metrics = [metric for metric in needed_metrics if metric not in found_surfs]
    if metrics:
        LOGGER.info(f'ANAT Stage 8: Creating GIFTI metrics for {metrics}')
        gifti_morph_wf = init_gifti_morphometrics_wf(morphometrics=metrics)
        ds_morph_wf = init_ds_surface_metrics_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            metrics=metrics,
            name='ds_morph_wf',
        )

        workflow.connect([
            (surface_recon_wf, gifti_morph_wf, [
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
            ]),
            (gifti_morph_wf, surfaces_buffer, [
                (f'outputnode.{metric}', metric) for metric in metrics
            ]),
            (sourcefile_buffer, ds_morph_wf, [('anat_source_files', 'inputnode.source_files')]),
            (gifti_morph_wf, ds_morph_wf, [
                (f'outputnode.{metric}', f'inputnode.{metric}') for metric in metrics
            ]),
        ])  # fmt:skip

    if 'anat_ribbon' not in precomputed:
        LOGGER.info('ANAT Stage 8a: Creating cortical ribbon mask')
        anat_ribbon_wf = init_anat_ribbon_wf()
        ds_ribbon_mask_wf = init_ds_mask_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            mask_type='ribbon',
            extra_entities={'space': reference_anat},
            name='ds_ribbon_mask_wf',
        )

        workflow.connect([
            (anat_buffer, anat_ribbon_wf, [
                ('anat_preproc', 'inputnode.ref_file'),
            ]),
            (surfaces_buffer, anat_ribbon_wf, [
                ('white', 'inputnode.white'),
                ('pial', 'inputnode.pial'),
            ]),
            (sourcefile_buffer, ds_ribbon_mask_wf, [
                ('anat_source_files', 'inputnode.source_files'),
            ]),
            (anat_ribbon_wf, ds_ribbon_mask_wf, [
                ('outputnode.anat_ribbon', 'inputnode.mask_file'),
            ]),
            (ds_ribbon_mask_wf, outputnode, [('outputnode.mask_file', 'anat_ribbon')]),
        ])  # fmt:skip
    else:
        LOGGER.info('ANAT Stage 8a: Found pre-computed cortical ribbon mask')
        outputnode.inputs.anat_ribbon = precomputed['anat_ribbon']

    # Stage 9: Baseline fsLR registration
    if recon_method == 'mcribs':
        if len(precomputed.get('sphere_reg_dhcpAsym', [])) < 2:
            LOGGER.info('ANAT Stage 9: Creating dhcp-fsLR registration sphere')
            fsLR_reg_wf = init_mcribs_dhcp_wf()

            ds_fsLR_reg_wf = init_ds_surfaces_wf(
                output_dir=output_dir,
                surfaces=['sphere_reg_dhcpAsym'],
                name='ds_fsLR_reg_wf',
            )

            workflow.connect([
                (surfaces_buffer, fsLR_reg_wf, [('sphere_reg', 'inputnode.sphere_reg')]),
                (sourcefile_buffer, ds_fsLR_reg_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (fsLR_reg_wf, ds_fsLR_reg_wf, [
                    ('outputnode.sphere_reg_dhcpAsym', 'inputnode.sphere_reg_dhcpAsym')
                ]),
                (ds_fsLR_reg_wf, fsLR_buffer, [
                    ('outputnode.sphere_reg_dhcpAsym', 'sphere_reg_fsLR'),
                ]),
            ])  # fmt:skip
        else:
            LOGGER.info('ANAT Stage 9: Found pre-computed dhcp-fsLR registration sphere')
            fsLR_buffer.inputs.sphere_reg_fsLR = sorted(precomputed['sphere_reg_dhcpAsym'])

    else:
        if len(precomputed.get('sphere_reg_fsLR', [])) < 2:
            LOGGER.info('ANAT Stage 9: Creating fsLR registration sphere')
            fsLR_reg_wf = init_fsLR_reg_wf()

            ds_fsLR_reg_wf = init_ds_surfaces_wf(
                output_dir=output_dir,
                surfaces=['sphere_reg_fsLR'],
                name='ds_fsLR_reg_wf',
            )

            workflow.connect([
                (surfaces_buffer, fsLR_reg_wf, [('sphere_reg', 'inputnode.sphere_reg')]),
                (sourcefile_buffer, ds_fsLR_reg_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (fsLR_reg_wf, ds_fsLR_reg_wf, [
                    ('outputnode.sphere_reg_fsLR', 'inputnode.sphere_reg_fsLR')
                ]),
                (ds_fsLR_reg_wf, fsLR_buffer, [('outputnode.sphere_reg_fsLR', 'sphere_reg_fsLR')]),
            ])  # fmt:skip
        else:
            LOGGER.info('ANAT Stage 9: Found pre-computed fsLR registration sphere')
            fsLR_buffer.inputs.sphere_reg_fsLR = sorted(precomputed['sphere_reg_fsLR'])
    return workflow


@tag('anat.fit')
def init_infant_single_anat_fit_wf(
    *,
    age_months: int,
    t1w: list,
    t2w: list,
    flair: list,
    bids_root: str,
    precomputed: dict,
    longitudinal: bool,
    omp_nthreads: int,
    output_dir: str,
    segmentation_atlases: str | Path | None,
    skull_strip_mode: ty.Literal['force', 'skip', 'auto'],
    skull_strip_template: 'Reference',
    skull_strip_fixed_seed: bool,
    sloppy: bool,
    spaces: 'SpatialReferences',
    recon_method: ty.Literal['freesurfer', 'infantfs', 'mcribs'] | None,
    reference_anat: ty.Literal['T1w', 'T2w'],
    cifti_output: ty.Literal['91k', '170k', False],
    msm_sulc: bool = False,
    name: str = 'infant_single_anat_fit_wf',
) -> Workflow:
    """
    Create a fit workflow with just a single anatomical.

    Note: Treat this functionality as a workaround that enables the processing of incomplete data.
    For best results, especially in periods of transitioning myelination (usually 3-8 months),
    a combination of T1w and T2w images will produce more accurate results.
    """
    anatomicals = t1w or t2w

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['t1w', 't2w', 'anat', 'roi', 'flair', 'subjects_dir', 'subject_id'],
        ),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                # Primary derivatives
                'anat_preproc',
                'anat_mask',
                'anat_dseg',
                'anat_tpms',
                'anat2std_xfm',
                'fsnative2anat_xfm',
                # Surface and metric derivatives for fsLR resampling
                'white',
                'pial',
                'midthickness',
                'sphere',
                'thickness',
                'sulc',
                'sphere_reg',
                'sphere_reg_fsLR',
                'sphere_reg_msm',
                'anat_ribbon',
                # Reverse transform; not computable from forward transform
                'std2anat_xfm',
                # Metadata
                'template',
                'subjects_dir',
                'subject_id',
                'anat_valid_list',
            ]
        ),
        name='outputnode',
    )

    anat = reference_anat.lower()
    workflow = Workflow(name=f'infant_single_{anat}_fit_wf')
    workflow.add_nodes([inputnode, outputnode])

    # Stage 1 inputs (filtered)
    sourcefile_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_source_files']),
        name='sourcefile_buffer',
    )

    # Stage 2
    anat_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_preproc', 'anat_mask', 'anat_brain']),
        name='anat_buffer',
    )

    # Additional buffer if CSF normalization is used
    anat_reg_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_preproc']),
        name='anat_reg_buffer',
    )
    if not config.workflow.norm_csf:
        workflow.connect(anat_buffer, 'anat_preproc', anat_reg_buffer, 'anat_preproc')

    aseg_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_aseg']),
        name='aseg_buffer',
    )

    # Stage 3 - Segmentation
    seg_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_dseg', 'anat_tpms', 'ants_segs']),
        name='seg_buffer',
    )
    # Stage 4 - collated template names, forward and reverse transforms
    template_buffer = pe.Node(niu.Merge(3), name='template_buffer')
    anat2std_buffer = pe.Node(niu.Merge(3), name='anat2std_buffer')
    std2anat_buffer = pe.Node(niu.Merge(3), name='std2anat_buffer')

    # Stage 5 results: Refined stage 2 results; may be direct copy if no refinement
    refined_buffer = pe.Node(
        niu.IdentityInterface(fields=['anat_mask', 'anat_brain']),
        name='refined_buffer',
    )

    fsnative_buffer = pe.Node(
        niu.IdentityInterface(fields=['fsnative2anat_xfm', 'anat2fsnative_xfm']),
        name='fsnative_buffer',
    )

    # Stage 6 results: GIFTI surfaces
    surfaces_buffer = pe.Node(
        niu.IdentityInterface(
            fields=['white', 'pial', 'midthickness', 'sphere', 'sphere_reg', 'thickness', 'sulc']
        ),
        name='surfaces_buffer',
    )

    # Stage 7 and 8 results: fsLR sphere registration
    fsLR_buffer = pe.Node(niu.IdentityInterface(fields=['sphere_reg_fsLR']), name='fsLR_buffer')
    msm_buffer = pe.Node(niu.IdentityInterface(fields=['sphere_reg_msm']), name='msm_buffer')

    workflow.connect([
        (anat_buffer, outputnode, [
            ('anat_preproc', 'anat_preproc'),
        ]),
        (refined_buffer, outputnode, [
            ('anat_mask', 'anat_mask'),
            ('anat_brain', 'anat_brain'),
        ]),
        (seg_buffer, outputnode, [
            ('anat_dseg', 'anat_dseg'),
            ('anat_tpms', 'anat_tpms'),
        ]),
        (anat2std_buffer, outputnode, [('out', 'anat2std_xfm')]),
        (std2anat_buffer, outputnode, [('out', 'std2anat_xfm')]),
        (template_buffer, outputnode, [('out', 'template')]),
        (sourcefile_buffer, outputnode, [('anat_source_files', 'anat_valid_list')]),
        (surfaces_buffer, outputnode, [
            ('white', 'white'),
            ('pial', 'pial'),
            ('midthickness', 'midthickness'),
            ('sphere', 'sphere'),
            ('sphere_reg', 'sphere_reg'),
            ('thickness', 'thickness'),
            ('sulc', 'sulc'),
        ]),
        (fsLR_buffer, outputnode, [('sphere_reg_fsLR', 'sphere_reg_fsLR')]),
        (msm_buffer, outputnode, [('sphere_reg_msm', 'sphere_reg_msm')]),
    ])  # fmt:skip

    # Reporting
    anat_reports_wf = init_anat_reports_wf(
        spaces=spaces,
        surface_recon=recon_method,
        output_dir=output_dir,
        sloppy=sloppy,
    )

    workflow.connect([
        (outputnode, anat_reports_wf, [
            ('anat_valid_list', 'inputnode.source_file'),
            ('anat_preproc', 'inputnode.anat_preproc'),
            ('anat_mask', 'inputnode.anat_mask'),
            ('anat_dseg', 'inputnode.anat_dseg'),
            ('template', 'inputnode.template'),
            ('anat2std_xfm', 'inputnode.anat2std_xfm'),
            ('subjects_dir', 'inputnode.subjects_dir'),
            ('subject_id', 'inputnode.subject_id'),
        ]),
    ])  # fmt:skip

    desc = (
        '\nAnatomical data preprocessing\n\n: '
        f'A total of {len(anatomicals)} {anat} images were found '
        'within the input BIDS dataset.\n'
    )
    # Lowercase to match pattern
    anat = reference_anat.lower()

    # Derivatives
    anat_preproc = precomputed.get(f'{anat}_preproc')
    anat_mask = precomputed.get(f'{anat}_mask')
    anat_tpms = precomputed.get(f'{anat}_tpms')
    anat_dseg = precomputed.get(f'{anat}_dseg')
    anat_aseg = precomputed.get(f'{anat}_aseg')

    anat_validate = pe.Node(ValidateImage(), name='anat_validate', run_without_submitting=True)
    if not anat_preproc:
        LOGGER.info(f'ANAT Stage 1: Adding {reference_anat} template workflow')
        desc += (
            f'The {reference_anat} image was denoised and corrected for intensity '
            'non-uniformity (INU)'
        )

        anat_template_wf = init_anat_template_wf(
            image_type=reference_anat,
            num_files=len(anatomicals),
            longitudinal=longitudinal,
            omp_nthreads=omp_nthreads,
            name=f'{anat}_template_wf',
        )
        ds_anat_template_wf = init_ds_template_wf(
            image_type=reference_anat,
            output_dir=output_dir,
            num_anat=len(anatomicals),
            name='ds_anat_template_wf',
        )

        workflow.connect([
            (anat_template_wf, sourcefile_buffer, [
                ('outputnode.anat_valid_list', 'anat_source_files'),
            ]),
            (anat_template_wf, anat_reports_wf, [
                ('outputnode.out_report', 'inputnode.anat_conform_report'),
            ]),
        ])  # fmt:skip

        workflow.connect([
            (inputnode, anat_template_wf, [('anat', 'inputnode.anat_files')]),
            (anat_template_wf, anat_validate, [('outputnode.anat_ref', 'in_file')]),
            (anat_template_wf, ds_anat_template_wf, [
                ('outputnode.anat_realign_xfm', 'inputnode.anat_ref_xfms'),
            ]),
            (sourcefile_buffer, ds_anat_template_wf, [
                ('anat_source_files', 'inputnode.source_files'),
            ]),
            (anat_buffer, ds_anat_template_wf, [('anat_preproc', 'inputnode.anat_preproc')]),
        ])  # fmt:skip
    else:
        LOGGER.info('ANAT Found preprocessed T1w - skipping Stage 1')
        desc += ' A preprocessed T1w image was provided as input.'

        anat_validate.inputs.in_file = anat_preproc
        sourcefile_buffer.inputs.anat_source_files = [anat_preproc]

        workflow.connect([
            (anat_validate, anat_buffer, [('out_file', 'anat_preproc')]),
        ])  # fmt:skip

    # Stage 2: Use previously computed mask or calculate
    # If we only have one mask (could be either T1w/T2w),
    # just apply transform to get it in the other space
    # T2w masking logic:
    # PCM = Pre-computed mask
    # PCM = Pre-computed mask for PCT
    # PCT = Pre-computed template
    # SS = Skull stripping required
    #
    # PCM, PCT, SS -> Apply PCM to PCT
    # PCM, PCT, !SS -> Apply PCM to PCT
    # PCM, !PCT, SS -> Apply PCM to template, then run INU
    # PCM, !PCT, !SS -> Apply PCM to template, then run INU
    # !PCM, PCT, SS -> Run brain extraction
    # !PCM, PCT, !SS -> Binarize PCT
    # !PCM, !PCT, SS -> Run brain extraction
    # !PCM, !PCT, !SS -> INU correct template
    refine_mask = False
    apply_mask = pe.Node(ApplyMask(), name='apply_mask')
    if not anat_mask:
        if skull_strip_mode == 'auto':
            run_skull_strip = not all(_is_skull_stripped(img) for img in anatomicals)
        else:
            run_skull_strip = {'force': True, 'skip': False}[skull_strip_mode]

        if not run_skull_strip:
            desc += (
                f'The {reference_anat} reference was previously skull-stripped; '
                'a brain mask was derived from the input image.'
            )

            if not anat_preproc:
                LOGGER.info('ANAT Stage 2b: Skipping skull-strip, INU-correction only')

                n4_only_wf = init_n4_only_wf(
                    omp_nthreads=omp_nthreads,
                    bids_suffix=reference_anat,
                    atropos_use_random_seed=not skull_strip_fixed_seed,
                    name='n4_only_wf',
                )

                refine_mask = True
                workflow.connect([
                    (anat_validate, n4_only_wf, [('out_file', 'inputnode.in_files')]),
                    (n4_only_wf, anat_buffer, [
                        (('outputnode.bias_corrected', pop_file), 'anat_preproc'),
                        ('outputnode.out_mask', 'anat_mask'),
                        (('outputnode.out_file', pop_file), 'anat_brain'),
                    ]),
                    (n4_only_wf, seg_buffer, [
                        ('outputnode.out_segm', 'ants_segs'),
                    ]),
                ])  # fmt:skip
            else:
                LOGGER.info('ANAT Stage 2b: Skipping skull-strip, generating mask from input')
                binarize = pe.Node(Binarize(thresh_low=2), name='binarize')
                workflow.connect([
                    (anat_validate, binarize, [('out_file', 'in_file')]),
                    (anat_validate, anat_buffer, [('out_file', 'anat_brain')]),
                    (binarize, anat_buffer, [('out_file', 'anat_mask')]),
                ])  # fmt:skip
        else:
            LOGGER.info(f'ANAT Atlas-based brain mask will be calculated on the {reference_anat}')
            brain_extraction_wf = init_infant_brain_extraction_wf(
                omp_nthreads=omp_nthreads,
                sloppy=sloppy,
                age_months=age_months,
                ants_affine_init=True,
                skull_strip_template=skull_strip_template.space,
                template_specs=skull_strip_template.spec,
                debug='registration' in config.execution.debug,
            )

            workflow.connect([
                (anat_validate, brain_extraction_wf, [
                    ('out_file', 'inputnode.t2w_preproc'),
                ]),
                (brain_extraction_wf, anat_buffer, [
                    ('outputnode.out_mask', 'anat_mask'),
                    ('outputnode.t2w_brain', 'anat_brain'),
                ]),
            ])  # fmt:skip

    else:
        LOGGER.info(f'ANAT Found {reference_anat} brain mask')
        desc += 'A pre-computed brain mask was provided as input and used throughout the workflow.'
        workflow.connect(apply_mask, 'out_file', anat_buffer, 'anat_brain')

        if not anat_preproc:
            conform_anat_mask_wf = init_conform_derivative_wf(
                in_file=anat_mask,
                name='conform_anat_mask_wf',
            )

            LOGGER.info('ANAT Skipping skull-strip, INU-correction only')
            anat_n4_wf = init_anat_preproc_wf(name='anat_n4_wf')
            workflow.connect([
                (anat_validate, conform_anat_mask_wf, [('out_file', 'inputnode.ref_file')]),
                (conform_anat_mask_wf, anat_buffer, [('outputnode.out_file', 'anat_mask')]),
                (conform_anat_mask_wf, apply_mask, [('outputnode.out_file', 'in_mask')]),
                (anat_validate, anat_n4_wf, [('out_file', 'inputnode.in_anat')]),
                (anat_n4_wf, anat_buffer, [('outputnode.anat_preproc', 'anat_preproc')]),
                (anat_n4_wf, apply_mask, [('outputnode.anat_preproc', 'in_file')]),
            ])  # fmt:skip
        else:
            LOGGER.info(f'ANAT Skipping {reference_anat} masking')
            workflow.connect(anat_validate, 'out_file', apply_mask, 'in_file')
            anat_buffer.inputs.anat_mask = anat_mask
            apply_mask.inputs.in_mask = anat_mask

    # Stage 3: Segmentation
    seg_method = 'jlf' if config.execution.segmentation_atlases_dir else 'fast'
    if anat_aseg:
        LOGGER.info('ANAT Found precomputed anatomical segmentation')
        # Ensure compatibility with anatomical template
        if not anat_preproc:
            conform_aseg_wf = init_conform_derivative_wf(
                in_file=anat_aseg,
                name='conform_aseg_wf',
            )

            workflow.connect([
                (anat_buffer, conform_aseg_wf, [('anat_preproc', 'inputnode.ref_file')]),
                (conform_aseg_wf, aseg_buffer, [('outputnode.out_file', 'anat_aseg')]),
            ])  # fmt:skip
        else:
            aseg_buffer.inputs.anat_aseg = anat_aseg

    if not (anat_dseg and anat_tpms):
        LOGGER.info('ANAT Stage 3: Tissue segmentation')
        segmentation_wf = init_segmentation_wf(
            sloppy=sloppy,
            method=seg_method,
            image_type=reference_anat,
            omp_nthreads=omp_nthreads,
            has_aseg=bool(anat_aseg),
        )

        workflow.connect([
            (anat_buffer, segmentation_wf, [('anat_brain', 'inputnode.anat_brain')]),
            (segmentation_wf, seg_buffer, [
                ('outputnode.anat_dseg', 'anat_dseg'),
                ('outputnode.anat_tpms', 'anat_tpms'),
            ]),
        ])  # fmt:skip

        if anat_aseg:
            workflow.connect(aseg_buffer, 'anat_aseg', segmentation_wf, 'inputnode.anat_aseg')
        elif seg_method == 'jlf':
            workflow.connect(segmentation_wf, 'outputnode.anat_aseg', aseg_buffer, 'anat_aseg')
            # TODO: datasink aseg

        if not anat_dseg:
            ds_dseg_wf = init_ds_dseg_wf(
                output_dir=str(output_dir),
                extra_entities={'space': reference_anat},
            )
            workflow.connect([
                (sourcefile_buffer, ds_dseg_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (segmentation_wf, ds_dseg_wf, [
                    ('outputnode.anat_dseg', 'inputnode.anat_dseg'),
                ]),
            ])  # fmt:skip

        if not anat_tpms:
            ds_tpms_wf = init_ds_tpms_wf(
                output_dir=str(output_dir),
                extra_entities={'space': reference_anat},
            )
            workflow.connect([
                (sourcefile_buffer, ds_tpms_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (segmentation_wf, ds_tpms_wf, [
                    ('outputnode.anat_tpms', 'inputnode.anat_tpms'),
                ]),
            ])  # fmt:skip
    else:
        LOGGER.info('ANAT Stage 3: Skipping segmentation workflow')
    if anat_dseg:
        LOGGER.info('ANAT Found discrete segmentation')
        desc += 'Precomputed discrete tissue segmentations were provided as inputs.\n'
        seg_buffer.inputs.anat_dseg = anat_dseg
    if anat_tpms:
        LOGGER.info('ANAT Found tissue probability maps')
        desc += 'Precomputed tissue probabiilty maps were provided as inputs.\n'
        seg_buffer.inputs.anat_tpms = anat_tpms

    # Stage 4: Normalization
    # If selected adult templates are requested (MNI152 6th Gen or 2009)
    # opt to concatenate transforms first from native -> infant template (MNIInfant),
    # and then use a previously computed MNIInfant<cohort> -> MNI transform
    # this minimizes the chance of a bad registration.

    templates = []
    concat_xfms = []
    found_xfms = {}
    intermediate = None  # The intermediate space when concatenating xfms - includes cohort
    intermediate_targets = (
        {
            'MNI152NLin6Asym',  # TODO: 'MNI152NLin2009cAsym'
        }
        if config.workflow.multi_step_reg
        else set()
    )

    for template in spaces.get_spaces(nonstandard=False, dim=(3,)):
        # resolution / spec will not differentiate here
        if template.startswith('MNIInfant'):
            intermediate = template
        xfms = precomputed.get('transforms', {}).get(template, {})
        if set(xfms) != {'forward', 'reverse'}:
            if template in intermediate_targets:
                concat_xfms.append(template)
            else:
                templates.append(template)
        else:
            found_xfms[template] = xfms

    # Create another set of buffers to handle the case where we aggregate found and generated
    # xfms to be concatenated
    concat_template_buffer = pe.Node(niu.Merge(2), name='concat_template_buffer')
    concat_template_buffer.inputs.in1 = list(found_xfms)
    concat_anat2std_buffer = pe.Node(niu.Merge(2), name='concat_anat2std_buffer')
    concat_anat2std_buffer.inputs.in1 = [xfm['forward'] for xfm in found_xfms.values()]
    concat_std2anat_buffer = pe.Node(niu.Merge(2), name='concat_std2anat_buffer')
    concat_std2anat_buffer.inputs.in1 = [xfm['reverse'] for xfm in found_xfms.values()]

    template_buffer.inputs.in1 = list(found_xfms)
    anat2std_buffer.inputs.in1 = [xfm['forward'] for xfm in found_xfms.values()]
    std2anat_buffer.inputs.in1 = [xfm['reverse'] for xfm in found_xfms.values()]

    if config.workflow.norm_csf:
        csf_norm_wf = init_csf_norm_wf()

        workflow.connect([
            (anat_buffer, csf_norm_wf, [('anat_preproc', 'inputnode.anat_preproc')]),
            (seg_buffer, csf_norm_wf, [('anat_tpms', 'inputnode.anat_tpms')]),
            (csf_norm_wf, anat_reg_buffer, [('outputnode.anat_preproc', 'anat_preproc')]),
        ])  # fmt:skip

    if templates:
        LOGGER.info(f'ANAT Stage 4: Preparing normalization workflow for {templates}')
        register_template_wf = init_register_template_wf(
            sloppy=sloppy,
            omp_nthreads=omp_nthreads,
            templates=templates,
            image_type=reference_anat,
        )
        ds_template_registration_wf = init_ds_template_registration_wf(
            output_dir=str(output_dir),
            image_type=reference_anat,
        )

        workflow.connect([
            (inputnode, register_template_wf, [('roi', 'inputnode.lesion_mask')]),
            (anat_reg_buffer, register_template_wf, [
                ('anat_preproc', 'inputnode.moving_image'),
            ]),
            (refined_buffer, register_template_wf, [('anat_mask', 'inputnode.moving_mask')]),
            (sourcefile_buffer, ds_template_registration_wf, [
                ('anat_source_files', 'inputnode.source_files')
            ]),
            (register_template_wf, ds_template_registration_wf, [
                ('outputnode.template', 'inputnode.template'),
                ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('outputnode.std2anat_xfm', 'inputnode.std2anat_xfm'),
            ]),
            (register_template_wf, template_buffer, [('outputnode.template', 'in2')]),
            (register_template_wf, concat_template_buffer, [('outputnode.template', 'in2')]),
            (register_template_wf, std2anat_buffer, [('outputnode.std2anat_xfm', 'in2')]),
            (register_template_wf, concat_std2anat_buffer, [('outputnode.std2anat_xfm', 'in2')]),
            (register_template_wf, anat2std_buffer, [('outputnode.anat2std_xfm', 'in2')]),
            (register_template_wf, concat_anat2std_buffer, [('outputnode.anat2std_xfm', 'in2')]),
        ])  # fmt:skip

    if concat_xfms:
        LOGGER.info(f'ANAT Stage 5b: Concatenating normalization for {concat_xfms}')

        select_infant_mni = pe.Node(
            KeySelect(fields=['anat2std_xfm', 'std2anat_xfm'], key=intermediate),
            name='select_infant_mni',
            run_without_submitting=True,
        )

        concat_reg_wf = init_concat_registrations_wf(templates=concat_xfms)
        ds_concat_reg_wf = init_ds_template_registration_wf(
            output_dir=str(output_dir),
            image_type=reference_anat,
            name='ds_concat_registration_wf',
        )

        workflow.connect([
            (concat_template_buffer, select_infant_mni, [('out', 'keys')]),
            (concat_anat2std_buffer, select_infant_mni, [('out', 'anat2std_xfm')]),
            (concat_std2anat_buffer, select_infant_mni, [('out', 'std2anat_xfm')]),
            (select_infant_mni, concat_reg_wf, [
                ('key', 'inputnode.intermediate'),
                ('anat2std_xfm', 'inputnode.anat2int_xfm'),
                ('std2anat_xfm', 'inputnode.int2anat_xfm'),
            ]),
            (sourcefile_buffer, ds_concat_reg_wf, [
                ('anat_source_files', 'inputnode.source_files')
            ]),
            (concat_reg_wf, ds_concat_reg_wf, [
                ('outputnode.template', 'inputnode.template'),
                ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('outputnode.std2anat_xfm', 'inputnode.std2anat_xfm'),
            ]),
            (concat_reg_wf, template_buffer, [('outputnode.template', 'in3')]),
            (concat_reg_wf, anat2std_buffer, [('outputnode.anat2std_xfm', 'in3')]),
            (concat_reg_wf, std2anat_buffer, [('outputnode.std2anat_xfm', 'in3')]),
        ])  # fmt:skip

    if found_xfms:
        LOGGER.info(f'ANAT Stage 4: Found pre-computed registrations for {found_xfms}')

    # Only refine mask if necessary
    if anat_mask or recon_method is None:
        workflow.connect([
            (anat_buffer, refined_buffer, [
                ('anat_mask', 'anat_mask'),
                ('anat_brain', 'anat_brain'),
            ]),
        ])  # fmt:skip

    workflow.__desc__ = desc

    if recon_method is None:
        LOGGER.info('ANAT Skipping Stages 5+')
        return workflow

    # Stage 5: Surface reconstruction
    if recon_method == 'mcribs':
        if reference_anat == 'T1w':
            LOGGER.warning('Attempting to use MCRIBS with a T1w file, good luck.')

        from nibabies.workflows.anatomical.surfaces import init_mcribs_surface_recon_wf

        LOGGER.info('ANAT Stage 5: Preparing M-CRIB-S reconstruction workflow')
        surface_recon_wf = init_mcribs_surface_recon_wf(
            omp_nthreads=omp_nthreads,
            use_aseg=bool(anat_aseg),
            precomputed=precomputed,
            mcribs_dir=str(config.execution.mcribs_dir),
        )

        workflow.connect([
            (inputnode, surface_recon_wf, [
                ('subject_id', 'inputnode.subject_id'),
                ('subjects_dir', 'inputnode.subjects_dir'),
            ]),
            (anat_validate, surface_recon_wf, [('out_file', 'inputnode.t2w')]),
            (anat_buffer, surface_recon_wf, [('anat_mask', 'inputnode.in_mask')]),
            (aseg_buffer, surface_recon_wf, [
                ('anat_aseg', 'inputnode.in_aseg'),
            ]),
            (surface_recon_wf, outputnode, [
                ('outputnode.subjects_dir', 'subjects_dir'),
                ('outputnode.subject_id', 'subject_id'),
            ]),
        ])  # fmt:skip

    else:
        from smriprep.utils.misc import fs_isRunning

        fs_isrunning = pe.Node(
            niu.Function(function=fs_isRunning), overwrite=True, name='fs_isrunning'
        )
        fs_isrunning.inputs.logger = LOGGER

        if recon_method == 'freesurfer':
            from smriprep.workflows.surfaces import init_surface_recon_wf

            LOGGER.info('ANAT Stage 5: Preparing FreeSurfer recon-all workflow')
            fs_isrunning = pe.Node(
                niu.Function(function=fs_isRunning), overwrite=True, name='fs_isrunning'
            )
            fs_isrunning.inputs.logger = LOGGER

            surface_recon_wf = init_surface_recon_wf(
                name='surface_recon_wf',
                omp_nthreads=omp_nthreads,
                hires=True,
                fs_no_resume=False,
                precomputed=precomputed,
            )

        elif recon_method == 'infantfs':
            from nibabies.workflows.anatomical.surfaces import init_infantfs_surface_recon_wf

            LOGGER.info('ANAT Stage 5: Preparing Infant FreeSurfer workflow')
            surface_recon_wf = init_infantfs_surface_recon_wf(
                age_months=age_months,
                precomputed=precomputed,
                omp_nthreads=omp_nthreads,
                use_aseg=bool(anat_aseg),
            )

        workflow.connect([
            (inputnode, fs_isrunning, [
                ('subjects_dir', 'subjects_dir'),
                ('subject_id', 'subject_id'),
            ]),
            (inputnode, surface_recon_wf, [
                ('subject_id', 'inputnode.subject_id'),
            ]),
            (fs_isrunning, surface_recon_wf, [('out', 'inputnode.subjects_dir')]),
            (anat_validate, surface_recon_wf, [('out_file', 'inputnode.t1w')]),
            (anat_buffer, surface_recon_wf, [('anat_brain', 'inputnode.skullstripped_t1')]),
            (surface_recon_wf, outputnode, [
                ('outputnode.subjects_dir', 'subjects_dir'),
                ('outputnode.subject_id', 'subject_id'),
            ]),
        ])  # fmt:skip

        if anat_aseg:
            workflow.connect(aseg_buffer, 'anat_aseg', surface_recon_wf, 'inputnode.in_aseg')

    fsnative_xfms = precomputed.get('transforms', {}).get('fsnative')
    if not fsnative_xfms:
        ds_fs_registration_wf = init_ds_fs_registration_wf(
            image_type=reference_anat, output_dir=output_dir
        )

        if recon_method == 'freesurfer':
            workflow.connect([
                (surface_recon_wf, fsnative_buffer, [
                    ('outputnode.fsnative2t1w_xfm', 'fsnative2anat_xfm'),
                    ('outputnode.t1w2fsnative_xfm', 'anat2fsnative_xfm'),
                ]),
            ])  # fmt:skip
        else:
            workflow.connect([
                (surface_recon_wf, fsnative_buffer, [
                    ('outputnode.fsnative2anat_xfm', 'fsnative2anat_xfm'),
                    ('outputnode.anat2fsnative_xfm', 'anat2fsnative_xfm'),
                ]),
            ])  # fmt:skip

        workflow.connect([
            (sourcefile_buffer, ds_fs_registration_wf, [
                ('anat_source_files', 'inputnode.source_files'),
            ]),
            (fsnative_buffer, ds_fs_registration_wf, [
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (fsnative_buffer, outputnode, [
                ('fsnative2anat_xfm', 'fsnative2anat_xfm'),
            ]),
        ])  # fmt:skip
    elif 'reverse' in fsnative_xfms:
        LOGGER.info('ANAT Found fsnative-to-anatomical transform - skipping registration')
        outputnode.inputs.fsnative2anat_xfm = fsnative_xfms['reverse']
    else:
        raise RuntimeError(
            'Found an anatomical-to-fsnative transform without the reverse. Time to handle this.'
        )

    if not anat_mask and refine_mask:
        LOGGER.info('ANAT Stage 6: Preparing mask refinement workflow')
        # Stage 6: Refine ANTs mask with FreeSurfer segmentation
        refinement_wf = init_refinement_wf()
        applyrefined = pe.Node(ApplyMask(), name='applyrefined')

        workflow.connect([
            (surface_recon_wf, refinement_wf, [
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
                ('outputnode.subject_id', 'inputnode.subject_id'),
            ]),
            (fsnative_buffer, refinement_wf, [
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (anat_buffer, refinement_wf, [
                ('anat_preproc', 'inputnode.reference_image'),
            ]),
            (seg_buffer, refinement_wf, [
                ('ants_segs', 'inputnode.ants_segs'),
            ]),
            (anat_buffer, applyrefined, [('anat_preproc', 'in_file')]),
            (refinement_wf, applyrefined, [('outputnode.out_brainmask', 'in_mask')]),
            (refinement_wf, refined_buffer, [('outputnode.out_brainmask', 'anat_mask')]),
            (applyrefined, refined_buffer, [('out_file', 'anat_brain')]),
        ])  # fmt:skip
    else:
        LOGGER.info('ANAT Found brain mask - skipping Stage 6')

    # Stages 7-9: Surface conversion and registration
    # sphere_reg is needed to generate sphere_reg_fsLR
    # sphere and sulc are needed to generate sphere_reg_msm
    # white, pial, midthickness and thickness are needed to resample in the cortical ribbon
    # TODO: Consider paring down or splitting into a subworkflow that can be called on-demand
    # A subworkflow would still need to check for precomputed outputs
    needed_anat_surfs = ['white', 'pial', 'midthickness']
    needed_metrics = ['thickness', 'sulc']
    needed_spheres = ['sphere_reg', 'sphere']

    # Detect pre-computed surfaces
    found_surfs = {
        surf: sorted(precomputed[surf])
        for surf in needed_anat_surfs + needed_metrics + needed_spheres
        if len(precomputed.get(surf, [])) == 2
    }
    if found_surfs:
        LOGGER.info(f'ANAT Stage 7: Found pre-converted surfaces for {list(found_surfs)}')
        surfaces_buffer.inputs.trait_set(**found_surfs)

    # Stage 8: Surface conversion
    surfs = [surf for surf in needed_anat_surfs if surf not in found_surfs]
    spheres = [sphere for sphere in needed_spheres if sphere not in found_surfs]
    if surfs or spheres:
        LOGGER.info(f'ANAT Stage 7: Creating GIFTI surfaces for {surfs + spheres}')
    if surfs:
        gifti_surfaces_wf = init_gifti_surfaces_wf(surfaces=surfs)
        ds_surfaces_wf = init_ds_surfaces_wf(output_dir=output_dir, surfaces=surfs)

        workflow.connect([
            (surface_recon_wf, gifti_surfaces_wf, [
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
            ]),
            (fsnative_buffer, gifti_surfaces_wf, [
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (gifti_surfaces_wf, surfaces_buffer, [
                (f'outputnode.{surf}', surf) for surf in surfs
            ]),
            (sourcefile_buffer, ds_surfaces_wf, [('anat_source_files', 'inputnode.source_files')]),
            (gifti_surfaces_wf, ds_surfaces_wf, [
                (f'outputnode.{surf}', f'inputnode.{surf}') for surf in surfs
            ]),
        ])  # fmt:skip
    if spheres:
        gifti_spheres_wf = init_gifti_surfaces_wf(
            surfaces=spheres, to_scanner=False, name='gifti_spheres_wf'
        )
        ds_spheres_wf = init_ds_surfaces_wf(
            output_dir=output_dir,
            surfaces=spheres,
            name='ds_spheres_wf',
        )

        workflow.connect([
            (surface_recon_wf, gifti_spheres_wf, [
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
                # No transform for spheres, following HCP pipelines' lead
            ]),
            (gifti_spheres_wf, surfaces_buffer, [
                (f'outputnode.{sphere}', sphere) for sphere in spheres
            ]),
            (sourcefile_buffer, ds_spheres_wf, [('anat_source_files', 'inputnode.source_files')]),
            (gifti_spheres_wf, ds_spheres_wf, [
                (f'outputnode.{sphere}', f'inputnode.{sphere}') for sphere in spheres
            ]),
        ])  # fmt:skip
    metrics = [metric for metric in needed_metrics if metric not in found_surfs]
    if metrics:
        LOGGER.info(f'ANAT Stage 8: Creating GIFTI metrics for {metrics}')
        gifti_morph_wf = init_gifti_morphometrics_wf(morphometrics=metrics)
        ds_morph_wf = init_ds_surface_metrics_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            metrics=metrics,
            name='ds_morph_wf',
        )

        workflow.connect([
            (surface_recon_wf, gifti_morph_wf, [
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
            ]),
            (gifti_morph_wf, surfaces_buffer, [
                (f'outputnode.{metric}', metric) for metric in metrics
            ]),
            (sourcefile_buffer, ds_morph_wf, [('anat_source_files', 'inputnode.source_files')]),
            (gifti_morph_wf, ds_morph_wf, [
                (f'outputnode.{metric}', f'inputnode.{metric}') for metric in metrics
            ]),
        ])  # fmt:skip

    if 'anat_ribbon' not in precomputed:
        LOGGER.info('ANAT Stage 8a: Creating cortical ribbon mask')
        anat_ribbon_wf = init_anat_ribbon_wf()
        ds_ribbon_mask_wf = init_ds_mask_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            mask_type='ribbon',
            extra_entities={'space': reference_anat},
            name='ds_ribbon_mask_wf',
        )

        workflow.connect([
            (anat_buffer, anat_ribbon_wf, [
                ('anat_preproc', 'inputnode.ref_file'),
            ]),
            (surfaces_buffer, anat_ribbon_wf, [
                ('white', 'inputnode.white'),
                ('pial', 'inputnode.pial'),
            ]),
            (sourcefile_buffer, ds_ribbon_mask_wf, [
                ('anat_source_files', 'inputnode.source_files'),
            ]),
            (anat_ribbon_wf, ds_ribbon_mask_wf, [
                ('outputnode.anat_ribbon', 'inputnode.mask_file'),
            ]),
            (ds_ribbon_mask_wf, outputnode, [('outputnode.mask_file', 'anat_ribbon')]),
        ])  # fmt:skip
    else:
        LOGGER.info('ANAT Stage 8a: Found pre-computed cortical ribbon mask')
        outputnode.inputs.anat_ribbon = precomputed['anat_ribbon']

    # Stage 9: Baseline fsLR registration
    if recon_method == 'mcribs':
        if len(precomputed.get('sphere_reg_dhcpAsym', [])) < 2:
            LOGGER.info('ANAT Stage 9: Creating dhcp-fsLR registration sphere')
            fsLR_reg_wf = init_mcribs_dhcp_wf()

            ds_fsLR_reg_wf = init_ds_surfaces_wf(
                output_dir=output_dir,
                surfaces=['sphere_reg_dhcpAsym'],
                name='ds_fsLR_reg_wf',
            )

            workflow.connect([
                (surfaces_buffer, fsLR_reg_wf, [('sphere_reg', 'inputnode.sphere_reg')]),
                (sourcefile_buffer, ds_fsLR_reg_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (fsLR_reg_wf, ds_fsLR_reg_wf, [
                    ('outputnode.sphere_reg_dhcpAsym', 'inputnode.sphere_reg_dhcpAsym')
                ]),
                (ds_fsLR_reg_wf, fsLR_buffer, [
                    ('outputnode.sphere_reg_dhcpAsym', 'sphere_reg_fsLR'),
                ]),
            ])  # fmt:skip
        else:
            LOGGER.info('ANAT Stage 9: Found pre-computed dhcp-fsLR registration sphere')
            fsLR_buffer.inputs.sphere_reg_fsLR = sorted(precomputed['sphere_reg_dhcpAsym'])

    else:
        if len(precomputed.get('sphere_reg_fsLR', [])) < 2:
            LOGGER.info('ANAT Stage 9: Creating fsLR registration sphere')
            fsLR_reg_wf = init_fsLR_reg_wf()

            ds_fsLR_reg_wf = init_ds_surfaces_wf(
                output_dir=output_dir,
                surfaces=['sphere_reg_fsLR'],
                name='ds_fsLR_reg_wf',
            )

            workflow.connect([
                (surfaces_buffer, fsLR_reg_wf, [('sphere_reg', 'inputnode.sphere_reg')]),
                (sourcefile_buffer, ds_fsLR_reg_wf, [
                    ('anat_source_files', 'inputnode.source_files'),
                ]),
                (fsLR_reg_wf, ds_fsLR_reg_wf, [
                    ('outputnode.sphere_reg_fsLR', 'inputnode.sphere_reg_fsLR')
                ]),
                (ds_fsLR_reg_wf, fsLR_buffer, [('outputnode.sphere_reg_fsLR', 'sphere_reg_fsLR')]),
            ])  # fmt:skip
        else:
            LOGGER.info('ANAT Stage 9: Found pre-computed fsLR registration sphere')
            fsLR_buffer.inputs.sphere_reg_fsLR = sorted(precomputed['sphere_reg_fsLR'])
    return workflow
